import copy
import warnings
import patsy
import numpy as np
import statsmodels.api as sm
import statsmodels.formula.api as smf
import matplotlib.pyplot as plt
from scipy.stats import logistic, norm

from zepid.causal.utils import propensity_score, stochastic_check_conditional
from zepid.causal.doublyrobust.utils import tmle_unit_bounds, tmle_unit_unbound
from zepid.calc import probability_to_odds, odds_to_probability, probability_bounds
from zepid.causal.utils import (exposure_machine_learner, outcome_machine_learner, stochastic_outcome_machine_learner,
                                stochastic_outcome_predict, missing_machine_learner, plot_kde, plot_love,
                                standardized_mean_differences, positivity, plot_kde_accuracy, outcome_accuracy,
                                check_input_data)


class TMLE:
    r"""Implementation of target maximum likelihood estimator. This implementation calculates TMLE for a
    time-fixed exposure and a single time-point outcome. By default standard parametric regression models are used to
    calculate the estimate of interest. The TMLE estimator allows users to instead use machine learning algorithms
    from sklearn and PyGAM.

    Note
    ----
    Valid confidence intervals are only attainable with certain machine learning algorithms. These algorithms must be
    Donsker class for valid confidence intervals. GAM and LASSO are examples of alogorithms that are Donsker class

    Note
    ----
    TMLE is a doubly-robust substitution estimator. TMLE obtains the target estimate in a single step. The
    single-step TMLE is described further by van der Laan. For further details, see the listed references.

    Continuous outcomes must be bounded between 0 and 1. TMLE does this automatically for the user. Additionally,
    the average treatment effect is estimate is back converted to the original scale of Y. When scaling Y as Y*,
    some values may take the value of 0 or 1, which breaks a logit(Y*) transformation. To avoid this issue, Y* is
    bounded by the `continuous_bound` argument. The default is 0.0005, the same as R's tmle

    The following is a general outline of the estimation process for TMLE

    1. Initial estimates for Y are predicted from a regression model. Expected values for each individual are
    generated under the scenarios of all treated vs all untreated

    .. math::

        E(Y|A, L)

    2. Predicted probabilities are generated from a regression model

    .. math::

        \pi_1 = \Pr(A=1|L)

    3. The 'clever covariate' is calculated by

    .. math::

        H_a(A=a,L) = \frac{I(A=1)}{\pi_1} - \frac{I(A=0)}{\pi_0}

    for each individual. Afterwards, the predicted Y is set as an offset in the following logit model and used to
    predict values under each treatment strategy after fitted

    .. math::

        \text{logit}(E(Y|A,L)) = \text{logit}(Y_a) + \sigma  H_a

    4. The targeted Psi is estimated, representing the causal effect of all treated vs. all untreated

    Confidence intervals are constructed using influence curves.

    Parameters
    ----------
    df : DataFrame
        Pandas dataframe containing the variables of interest
    exposure : str
        Column label for the exposure of interest
    outcome : str
        Column label for the outcome of interest
    alpha : float, optional
        Alpha for confidence interval level. Default is 0.05
    continuous_bound : float, optional
        Optional argument to control the bounding feature for continuous outcomes. The bounding process may result
        in values of 0,1 which are undefined for logit(x). This parameter adds or substracts from the scenarios of
        0,1 respectively. Default value is 0.0005


    Examples
    --------
    Setting up environment

    >>> from zepid import load_sample_data, spline
    >>> from zepid.causal.doublyrobust import TMLE
    >>> df = load_sample_data(False).dropna()
    >>> df[['cd4_rs1', 'cd4_rs2']] = spline(df, 'cd40', n_knots=3, term=2, restricted=True)

    Estimating TMLE using logistic regression

    >>> tmle = TMLE(df, exposure='art', outcome='dead')
    >>> # Specifying exposure/treatment model
    >>> tmle.exposure_model('male + age0 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
    >>> # Specifying outcome model
    >>> tmle.outcome_model('art + male + age0 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
    >>> # TMLE estimation procedure
    >>> tmle.fit()
    >>> # Printing main results
    >>> tmle.summary()
    >>> # Extracting risk difference and confidence intervals, respectively
    >>> tmle.risk_difference
    >>> tmle.risk_difference_ci

    Estimating TMLE with machine learning algorithm from sklearn

    >>> from sklearn.linear_model import LogisticRegression
    >>> log1 = LogisticRegression(penalty='l1', random_state=201)
    >>> tmle = TMLE(df, 'art', 'dead')
    >>> # custom_model allows specification of machine learning algorithms
    >>> tmle.exposure_model('male + age0 + cd40 + cd4_rs1 + cd4_rs2 + dvl0', custom_model=log1)
    >>> tmle.outcome_model('male + age0 + cd40 + cd4_rs1 + cd4_rs2 + dvl0', custom_model=log1)
    >>> tmle.fit()

    Demonstration of estimating g-model with symmetric bounds

    >>> tmle.exposure_model('male + age0 + cd40 + cd4_rs1 + cd4_rs2 + dvl0', bound=0.05)

    Demonstration of estimating g-model with asymmetric bounds

    >>> tmle.exposure_model('male + age0 + cd40 + cd4_rs1 + cd4_rs2 + dvl0', bound=[0.05, 0.9])

    References
    ----------
    Schuler MS, and Sherri R. "Targeted maximum likelihood estimation for causal inference in
    observational studies." American journal of epidemiology 185.1 (2017): 65-73.

    Van der Laan, MJ, and Sherri R. Targeted learning: causal inference for observational and experimental
    data. Springer Science & Business Media, 2011.

    Van Der Laan, MJ, Rubin D. "Targeted maximum likelihood learning." The International Journal of
    Biostatistics 2.1 (2006).

    Gruber S, van der Laan, MJ. (2011). tmle: An R package for targeted maximum likelihood estimation.
    """
    def __init__(self, df, exposure, outcome, alpha=0.05, continuous_bound=0.0005):
        self.exposure = exposure
        self.outcome = outcome
        self._missing_indicator = '__missing_indicator__'
        self.df, self._miss_flag, self._continuous_outcome = check_input_data(data=df,
                                                                              exposure=exposure,
                                                                              outcome=outcome,
                                                                              estimator="TMLE",
                                                                              drop_censoring=False,
                                                                              drop_missing=True,
                                                                              binary_exposure_only=True)

        # Detailed steps follow "Targeted Learning" chapter 4, figure 4.2 by van der Laan, Rose
        if self._continuous_outcome:
            self._continuous_min = np.min(self.df[outcome])
            self._continuous_max = np.max(self.df[outcome])
            self._cb = continuous_bound
            self.df[outcome] = tmle_unit_bounds(y=self.df[outcome], mini=self._continuous_min,
                                                maxi=self._continuous_max, bound=self._cb)
        else:
            self._cb = 0.0

        self._out_model = None
        self._exp_model = None
        self._miss_model = None
        self._out_model_custom = False
        self._exp_model_custom = False
        self._miss_model_custom = False
        self._fit_exposure_model = False
        self._fit_outcome_model = False
        self._fit_missing_model = False
        self.alpha = alpha

        self.QA0W = None
        self.QA1W = None
        self.QAW = None
        self.g1W = None
        self.g0W = None
        self.m1W = None
        self.m0W = None
        self._epsilon = None

        self.risk_difference = None
        self.risk_difference_ci = None
        self.risk_difference_se = None
        self.risk_ratio = None
        self.risk_ratio_ci = None
        self.risk_ratio_se = None
        self.odds_ratio = None
        self.odds_ratio_ci = None
        self.odds_ratio_se = None
        self.average_treatment_effect = None
        self.average_treatment_effect_ci = None
        self.average_treatment_effect_se = None

    def exposure_model(self, model, custom_model=None, bound=False, print_results=True):
        """Estimation of Pr(A=1|L), which is termed as g(A=1|L) in the literature

        Parameters
        ----------
        model : str
            Independent variables to predict the exposure. Example) 'var1 + var2 + var3'
        custom_model : optional
            Input for a custom model that is used in place of the logit model (default). The model must have the
            "fit()" and  "predict()" attributes. SciKit-Learn style models are supported as custom models. In the
            background, TMLE will fit the custom model and generate the predicted probablities
        bound : float, list, optional
            Value between 0,1 to truncate predicted probabilities. Helps to avoid near positivity violations.
            Specifying this argument can improve finite sample performance for random positivity violations. However,
            truncating weights leads to additional confounding. Default is False, meaning no truncation of
            predicted probabilities occurs. Providing a single float assumes symmetric trunctation, where values below
            or above the threshold are set to the threshold value. Alternatively a list of floats can be provided for
            asymmetric trunctation, with the first value being the lower bound and the second being the upper bound
        print_results : bool, optional
            Whether to print the fitted model results. Default is True (prints results)
        """
        self._exp_model = self.exposure + ' ~ ' + model
        self.__mweight = model

        # Step 3) Estimation of g-model (exposure model)
        if custom_model is None:
            fitmodel = propensity_score(self.df, self._exp_model, print_results=print_results)
            self.g1W = fitmodel.predict(self.df)

        # User-specified prediction model
        else:
            self._exp_model_custom = True
            data = patsy.dmatrix(model + ' - 1', self.df)
            self.g1W = exposure_machine_learner(xdata=np.asarray(data),
                                                ydata=np.asarray(self.df[self.exposure]),
                                                ml_model=copy.deepcopy(custom_model),
                                                print_results=print_results)

        self.g0W = 1 - self.g1W
        if bound:  # Bounding predicted probabilities if requested
            self.g1W = probability_bounds(self.g1W, bounds=bound)
            self.g0W = probability_bounds(self.g0W, bounds=bound)

        self._fit_exposure_model = True

    def missing_model(self, model, custom_model=None, bound=False, print_results=True):
        """Estimation of Pr(M=1|A,L), which is the missing data mechanism for the outcome. The corresponding observation
        probabilities are used to update the clever covariates for estimation of Qn.

        The initial estimate of Q is still based on complete observations only

        Parameters
        ----------
        model : str
            Independent variables to predict the exposure. Example) 'var1 + var2 + var3'. The treatment must be
            included for the missing data model
        custom_model : optional
            Input for a custom model that is used in place of the logit model (default). The model must have the
            "fit()" and  "predict()" attributes. Both sklearn and supylearner are supported as custom models. In the
            background, TMLE will fit the custom model and generate the predicted probablities
        bound: float, list, optional
            Value between 0,1 to truncate predicted probabilities. Helps to avoid near positivity violations.
            Specifying this argument can improve finite sample performance for random positivity violations. However,
            truncating weights leads to additional confounding. Default is False, meaning no truncation of
            predicted probabilities occurs. Providing a single float assumes symmetric trunctation, where values below
            or above the threshold are set to the threshold value. Alternatively a list of floats can be provided for
            asymmetric trunctation, with the first value being the lower bound and the second being the upper bound
        print_results : bool, optional
            Whether to print the fitted model results. Default is True (prints results)
        """
        # Error if no missing outcome data
        if not self._miss_flag:
            raise ValueError("No missing outcome data is present in the data set")

        # Warning if exposure is not included in the missingness of outcome model
        if self.exposure not in model:
            warnings.warn("For the specified missing outcome model, the exposure variable should be included in the "
                          "model", UserWarning)

        self._miss_model = self._missing_indicator + ' ~ ' + model

        # Step 3b) Prediction for M if missing outcome data exists
        if custom_model is None:  # Logistic Regression model for predictions
            fitmodel = propensity_score(self.df, self._miss_model, print_results=print_results)
            dfx = self.df.copy()
            dfx[self.exposure] = 1
            self.m1W = fitmodel.predict(dfx)
            dfx = self.df.copy()
            dfx[self.exposure] = 0
            self.m0W = fitmodel.predict(dfx)

        # User-specified model
        else:
            self._miss_model_custom = True
            data = patsy.dmatrix(model + ' - 1', self.df)

            dfx = self.df.copy()
            dfx[self.exposure] = 1
            adata = patsy.dmatrix(model + ' - 1', dfx)
            dfx = self.df.copy()
            dfx[self.exposure] = 0
            ndata = patsy.dmatrix(model + ' - 1', dfx)

            self.m1W, self.m0W = missing_machine_learner(xdata=np.array(data),
                                                         mdata=self.df[self._missing_indicator],
                                                         all_a=adata, none_a=ndata,
                                                         ml_model=copy.deepcopy(custom_model),
                                                         print_results=print_results)

        if bound:  # Bounding predicted probabilities if requested
            self.m1W = probability_bounds(self.m1W, bounds=bound)
            self.m0W = probability_bounds(self.m0W, bounds=bound)

        self._fit_missing_model = True

    def outcome_model(self, model, custom_model=None, bound=False, print_results=True,
                      continuous_distribution='gaussian'):
        """Estimation of E(Y|A,L,M=1), which is also written sometimes as Q(A,W,M=1) or Pr(Y=1|A,W,M=1). Estimation
        of this model is based on complete observations of Y only

        Parameters
        ----------
        model : str
            Independent variables to predict the exposure. Example) 'var1 + var2 + var3'
        custom_model : optional
            Input for a custom model that is used in place of the logit model (default). The model must have the
            "fit()" and  "predict()" attributes. Both sklearn and supylearner are supported as custom models. In the
            background, TMLE will fit the custom model and generate the predicted values
        bound : bool, optional
            This argument should ONLY be used if the outcome is continuous. Value between 0,1 to truncate the bounded
            predicted outcomes. Default is `False`, meaning no truncation of predicted outcomes occurs (unless a
            predicted outcome is outside the bounded continuous outcome). Providing a single float assumes symmetric
            trunctation. A list of floats can be provided for asymmetric trunctation.
        print_results : bool, optional
            Whether to print the fitted model results. Default is True (prints results)
        continuous_distribution : str, optional
            Distribution to use for continuous outcomes. Options are 'gaussian' for normal distributions and 'poisson'
            for Poisson distributions
        """
        if self.exposure not in model:
            warnings.warn("It looks like '" + self.exposure + "' is not included in the outcome model.")

        self._out_model = self.outcome + ' ~ ' + model

        if self._miss_flag:
            cc = self.df.copy().dropna()
        else:
            cc = self.df.copy()

        # Step 1) Prediction for Q (estimation of Q-model)
        if custom_model is None:  # Logistic Regression model for predictions
            self._continuous_type = continuous_distribution
            if self._continuous_outcome:
                if (continuous_distribution == 'gaussian') or (continuous_distribution == 'normal'):
                    f = sm.families.family.Gaussian()
                elif continuous_distribution == 'poisson':
                    f = sm.families.family.Poisson()
                else:
                    raise ValueError("Only 'gaussian' and 'poisson' distributions are supported")
                log = smf.glm(self._out_model, cc, family=f).fit()
            else:
                f = sm.families.family.Binomial()
                log = smf.glm(self._out_model, cc, family=f).fit()

            if print_results:
                print('==============================================================================')
                print('Outcome Model')
                print(log.summary())
                print('==============================================================================')

            # Step 2) Estimation under the scenarios
            dfx = self.df.copy()
            dfx[self.exposure] = 1
            self.QA1W = log.predict(dfx)
            dfx = self.df.copy()
            dfx[self.exposure] = 0
            self.QA0W = log.predict(dfx)

        # User-specified model
        else:
            self._out_model_custom = True
            data = patsy.dmatrix(model + ' - 1', cc)

            dfx = self.df.copy()
            dfx[self.exposure] = 1
            adata = patsy.dmatrix(model + ' - 1', dfx)
            dfx = self.df.copy()
            dfx[self.exposure] = 0
            ndata = patsy.dmatrix(model + ' - 1', dfx)

            self.QA1W, self.QA0W = outcome_machine_learner(xdata=np.asarray(data),
                                                           ydata=np.asarray(cc[self.outcome]),
                                                           all_a=adata, none_a=ndata,
                                                           ml_model=copy.deepcopy(custom_model),
                                                           continuous=self._continuous_outcome,
                                                           print_results=print_results)

        if not bound:  # Bounding predicted probabilities if requested
            bound = self._cb

        # This bounding step prevents continuous outcomes from being outside the range
        self.QA1W = probability_bounds(self.QA1W, bounds=bound)
        self.QA0W = probability_bounds(self.QA0W, bounds=bound)
        self.QAW = self.QA1W * self.df[self.exposure] + self.QA0W * (1 - self.df[self.exposure])
        self._fit_outcome_model = True

    def fit(self):
        """Calculate the effect measures from the predicted exposure probabilities and predicted outcome values using
        the TMLE procedure. Confidence intervals are calculated using influence curves.

        Note
        ----
        Exposure and outcome models must be specified prior to `fit()`

        Returns
        -------
        TMLE gains `risk_difference`, `risk_ratio`, and `odds_ratio` for binary outcomes and
        `average _treatment_effect` for continuous outcomes
        """
        if (self._fit_exposure_model is False) or (self._fit_outcome_model is False):
            raise ValueError('The exposure and outcome models must be specified before the psi estimate can '
                             'be generated')
        if self._miss_flag and not self._fit_missing_model:
            warnings.warn("No missing data model has been specified. All missing outcome data is assumed to be "
                          "missing completely at random. To relax this assumption to outcome data is missing at random"
                          "please use the `missing_model()` function", UserWarning)

        # Step 4) Calculating clever covariate (HAW)
        if self._miss_flag and self._fit_missing_model:
            self.g1W_total = self.g1W * self.m1W
            self.g0W_total = self.g0W * self.m0W
        else:
            self.g1W_total = self.g1W
            self.g0W_total = self.g0W
        H1W = self.df[self.exposure] / self.g1W_total
        H0W = -(1 - self.df[self.exposure]) / self.g0W_total
        HAW = H1W + H0W

        # Step 5) Estimating TMLE
        f = sm.families.family.Binomial()
        y = self.df[self.outcome]
        log = sm.GLM(y, np.column_stack((H1W, H0W)), offset=np.log(probability_to_odds(self.QAW)),
                     family=f, missing='drop').fit()
        self._epsilon = log.params
        Qstar1 = logistic.cdf(np.log(probability_to_odds(self.QA1W)) + self._epsilon[0] / self.g1W_total)
        Qstar0 = logistic.cdf(np.log(probability_to_odds(self.QA0W)) - self._epsilon[1] / self.g0W_total)
        Qstar = log.predict(np.column_stack((H1W, H0W)), offset=np.log(probability_to_odds(self.QAW)))

        # Step 6) Calculating Psi
        if self.alpha == 0.05:  # Without this, won't match R exactly. R relies on 1.96, while I use SciPy
            zalpha = 1.96
        else:
            zalpha = norm.ppf(1 - self.alpha / 2, loc=0, scale=1)

        # p-values are not implemented (doing my part to enforce CL over p-values)
        delta = np.where(self.df[self._missing_indicator] == 1, 1, 0)
        if self._continuous_outcome:
            # Calculating Average Treatment Effect
            Qstar = tmle_unit_unbound(Qstar, mini=self._continuous_min, maxi=self._continuous_max)
            Qstar1 = tmle_unit_unbound(Qstar1, mini=self._continuous_min, maxi=self._continuous_max)
            Qstar0 = tmle_unit_unbound(Qstar0, mini=self._continuous_min, maxi=self._continuous_max)

            self.average_treatment_effect = np.nanmean(Qstar1 - Qstar0)
            # Influence Curve for CL
            y_unbound = tmle_unit_unbound(self.df[self.outcome], mini=self._continuous_min, maxi=self._continuous_max)
            ic = np.where(delta == 1,
                          HAW * (y_unbound - Qstar) + (Qstar1 - Qstar0) - self.average_treatment_effect,
                          Qstar1 - Qstar0 - self.average_treatment_effect)
            seIC = np.sqrt(np.nanvar(ic, ddof=1) / self.df.shape[0])
            self.average_treatment_effect_se = seIC
            self.average_treatment_effect_ci = [self.average_treatment_effect - zalpha * seIC,
                                                self.average_treatment_effect + zalpha * seIC]
        else:
            # Calculating Risk Difference
            self.risk_difference = np.nanmean(Qstar1 - Qstar0)
            # Influence Curve for CL
            ic = np.where(delta == 1,
                          HAW * (self.df[self.outcome] - Qstar) + (Qstar1 - Qstar0) - self.risk_difference,
                          (Qstar1 - Qstar0) - self.risk_difference)
            seIC = np.sqrt(np.nanvar(ic, ddof=1) / self.df.shape[0])
            self.risk_difference_se = seIC
            self.risk_difference_ci = [self.risk_difference - zalpha * seIC,
                                       self.risk_difference + zalpha * seIC]

            # Calculating Risk Ratio
            self.risk_ratio = np.nanmean(Qstar1) / np.nanmean(Qstar0)
            # Influence Curve for CL
            ic = np.where(delta == 1,
                          (1 / np.mean(Qstar1) * (H1W * (self.df[self.outcome] - Qstar) + Qstar1 - np.mean(Qstar1)) -
                           (1/np.mean(Qstar0)) * (-1 * H0W * (self.df[self.outcome] - Qstar) + Qstar0 - np.mean(Qstar0))),
                          (Qstar1 - np.mean(Qstar1)) + Qstar0 - np.mean(Qstar0))

            seIC = np.sqrt(np.nanvar(ic, ddof=1) / self.df.shape[0])
            self.risk_ratio_se = seIC
            self.risk_ratio_ci = [np.exp(np.log(self.risk_ratio) - zalpha * seIC),
                                  np.exp(np.log(self.risk_ratio) + zalpha * seIC)]

            # Calculating Odds Ratio
            self.odds_ratio = (np.nanmean(Qstar1) / (1 - np.nanmean(Qstar1)
                                                     )) / (np.nanmean(Qstar0) / (1 - np.nanmean(Qstar0)))
            # Influence Curve for CL
            ic = np.where(delta == 1,
                          ((1 / (np.nanmean(Qstar1)*(1 - np.nanmean(Qstar1))) *
                            (H1W * (self.df[self.outcome] - Qstar) + Qstar1)) -
                           (1 / (np.nanmean(Qstar0)*(1 - np.nanmean(Qstar0))) *
                            (-1 * H0W * (self.df[self.outcome] - Qstar) + Qstar0))),

                          ((1 / (np.nanmean(Qstar1) * (1 - np.nanmean(Qstar1))) * Qstar1 -
                           (1 / (np.nanmean(Qstar0) * (1 - np.nanmean(Qstar0))) * Qstar0))))
            seIC = np.sqrt(np.nanvar(ic, ddof=1) / self.df.shape[0])
            self.odds_ratio_se = seIC
            self.odds_ratio_ci = [np.exp(np.log(self.odds_ratio) - zalpha * seIC),
                                  np.exp(np.log(self.odds_ratio) + zalpha * seIC)]

    def summary(self, decimal=3):
        """Prints summary of the estimated average causal effects

        Parameters
        ----------
        decimal : int, optional
            Number of decimal places to display. Default is 3
        """
        if (self._fit_exposure_model is False) or (self._fit_exposure_model is False):
            raise ValueError('The exposure and outcome models must be specified before the psi estimate can '
                             'be generated')

        print('======================================================================')
        print('                Targeted Maximum Likelihood Estimator                 ')
        print('======================================================================')
        fmt = 'Treatment:        {:<15} No. Observations:     {:<20}'
        print(fmt.format(self.exposure, self.df.shape[0]))

        fmt = 'Outcome:          {:<15} No. Missing Outcome:  {:<20}'
        print(fmt.format(self.outcome, np.sum(self.df[self.outcome].isnull())))

        fmt = 'g-Model:          {:<15} Missing Model:        {:<20}'
        if self._exp_model_custom:
            e = 'User-specified'
        else:
            e = 'Logistic'
        if self._miss_model_custom and self._miss_model is not None:
            m = 'User-specified'
        elif self._miss_model is None:
            m = 'None'
        else:
            m = 'Logistic'

        print(fmt.format(e, m))

        fmt = 'Q-Model:          {:<15}'
        if self._out_model_custom:
            y = 'User-specified'
        elif self._continuous_outcome:
            y = self._continuous_type
        else:
            y = 'Logistic'
        print(fmt.format(y))

        print('======================================================================')

        if self._continuous_outcome:
            print('Average Treatment Effect: ', round(float(self.average_treatment_effect), decimal))
            print(str(round(100 * (1 - self.alpha), 1)) + '% two-sided CI: (' +
                  str(round(self.average_treatment_effect_ci[0], decimal)), ',',
                  str(round(self.average_treatment_effect_ci[1], decimal)) + ')')
        else:
            print('Risk Difference:    ', round(float(self.risk_difference), decimal))
            print(str(round(100 * (1 - self.alpha), 1)) + '% two-sided CI: (' +
                  str(round(self.risk_difference_ci[0], decimal)), ',',
                  str(round(self.risk_difference_ci[1], decimal)) + ')')
            print('----------------------------------------------------------------------')
            print('Risk Ratio:         ', round(float(self.risk_ratio), decimal))
            print(str(round(100 * (1 - self.alpha), 1)) + '% two-sided CI: (' +
                  str(round(self.risk_ratio_ci[0], decimal)), ',',
                  str(round(self.risk_ratio_ci[1], decimal)) + ')')
            print('----------------------------------------------------------------------')
            print('Odds Ratio:         ', round(float(self.odds_ratio), decimal))
            print(str(round(100 * (1 - self.alpha), 1)) + '% two-sided CI: (' +
                  str(round(self.odds_ratio_ci[0], decimal)), ',',
                  str(round(self.odds_ratio_ci[1], decimal)) + ')')
        print('======================================================================')

    def run_diagnostics(self, decimal=3):
        """Run all currently implemented diagnostics for the exposure and outcome models. Each
        `run_diagnostics` provides results for all implemented diagnostics for ease of the user. For publication
        quality presentations, I recommend calling each diagnostic function individually and utilizing the optional
        parameters

        Note
        ----
        The plot presented cannot be edited. To edit the plots, call `plot_kde` or `plot_love` directly. Those
        functions return an axes object

        Parameters
        ----------
        decimal : int, optional
            Number of decimal places to display. Default is 3

        Returns
        -------
        None
        """
        if not self._fit_outcome_model or not self._fit_exposure_model:
            raise ValueError("The exposure_model and outcome_model function must be ran before any diagnostics")

        if self._fit_missing_model:
            ps = self.g1W * np.where(self.df[self.exposure] == 1, self.m1W, self.m0W)
        else:
            ps = self.g1W

        # Weight diagnostics
        print('\tExposure Model Diagnostics')
        self.positivity(decimal=decimal)

        print('======================================================================')
        print('                  Standardized Mean Differences')
        print('======================================================================')
        print(self.standardized_mean_differences().set_index(keys='labels'))
        print('======================================================================\n')

        # Outcome accuracy diagnostics
        print('\tOutcome Model Diagnostics')
        v = self.QAW - self.df[self.outcome]
        outcome_accuracy(true=self.df[self.outcome], predicted=self.QAW, decimal=decimal)

        df = self.df.copy()
        df['_ipw_'] = np.where(df[self.exposure] == 1, 1 / ps, 1 / (1 - ps))
        df['_g1_'] = ps

        plt.figure(figsize=[8, 6])
        plt.subplot(221)
        plot_love(df=df, treatment=self.exposure, weight='_ipw_', formula=self.__mweight)
        plt.title("Love Plot")

        plt.subplot(223)
        plot_kde(df=df, treatment=self.exposure, probability='_g1_')
        plt.title("Kernel Density of Propensity Scores")

        plt.subplot(222)
        plot_kde_accuracy(values=v.dropna(), color='green')
        plt.title("Kernel Density of Accuracy")
        plt.tight_layout()
        plt.show()

    def positivity(self, decimal=3):
        """Use this to assess whether positivity is a valid assumption for the exposure model / calculated IPTW. If
        there are extreme outliers, this may indicate problems with the calculated weights. To reduce extreme weights,
        the `bound` argument can be specified in `exposure_model()`

        Parameters
        --------------
        decimal : int, optional
            Number of decimal places to display. Default is three

        Returns
        --------------
        None
            Prints the positivity results to the console but does not return any objects
        """
        if self._fit_missing_model:
            ps = self.g1W * np.where(self.df[self.exposure] == 1, self.m1W, self.m0W)
        else:
            ps = self.g1W

        df = self.df.copy()
        df['_ipw_'] = np.where(df[self.exposure] == 1, 1 / ps, 1 / (1 - ps))
        pos = positivity(df=df, weights='_ipw_')
        print('======================================================================')
        print('                   Weight Positivity Diagnostics')
        print('======================================================================')
        print('If the mean of the weights is far from either the min or max, this may\n '
              'indicate the model is incorrect or positivity is violated')
        print('Average weight should be 2')
        print('----------------------------------------------------------------------')
        print('Mean weight:           ', round(pos[0], decimal))
        print('Standard Deviation:    ', round(pos[1], decimal))
        print('Minimum weight:        ', round(pos[2], decimal))
        print('Maximum weight:        ', round(pos[3], decimal))
        print('======================================================================\n')

    def standardized_mean_differences(self):
        """Calculates the standardized mean differences for all variables based on the inverse probability weights.

        Returns
        -------
        DataFrame
            Returns pandas DataFrame of calculated standardized mean differences. Columns are labels (variables labels),
            smd_u (unweighted standardized difference), and smd_w (weighted standardized difference)
        """
        if self._fit_missing_model:
            ps = self.g1W * np.where(self.df[self.exposure] == 1, self.m1W, self.m0W)
        else:
            ps = self.g1W

        df = self.df.copy()
        df['_ipw_'] = np.where(df[self.exposure] == 1, 1 / ps, 1 / (1 - ps))
        s = standardized_mean_differences(df=df, treatment=self.exposure,
                                          weight='_ipw_', formula=self.__mweight)
        return s

    def plot_kde(self, to_plot, bw_method='scott', fill=True,
                 color='g', color_e='b', color_u='r'):
        """Generates density plots that can be used to check predictions qualitatively. Density plots can be generated
        for assess either positivity violations of the exposure model or the accuracy in predicting the outcome for
        the outcome model. The kernel density used is SciPy's Gaussian kernel. Either Scott's Rule or
        Silverman's Rule can be implemented.

        Parameters
        ------------
        to_plot : str, optional
            The plot to generate. Specifying 'exposure' returns only the density plot for treatment probabilities,
            and 'outcome' returns only the density plot for the outcome accuracy
        bw_method : str, optional
            Method used to estimate the bandwidth. Following SciPy, either 'scott' or 'silverman' are valid options
        fill : bool, optional
            Whether to color the area under the density curves. Default is true
        color : str, optional
            Color of the line/area for predicted outcomes minus observed outcomes. Default is Green
        color_e : str, optional
            Color of the line/area for the treated group. Default is Blue
        color_u : str, optional
            Color of the line/area for the treated group. Default is Red

        Returns
        ---------------
        matplotlib axes
        """
        if not self._fit_outcome_model or not self._fit_exposure_model:
            raise ValueError("The exposure_model and outcome_model function must be ran before any diagnostics")

        if to_plot == 'exposure':
            if self._fit_missing_model:
                ps = self.g1W * np.where(self.df[self.exposure] == 1, self.m1W, self.m0W)
            else:
                ps = self.g1W
            df = self.df.copy()
            df['_g1_'] = ps
            ax = plot_kde(df=df, treatment=self.exposure, probability='_g1_',
                          bw_method=bw_method, fill=fill, color_e=color_e, color_u=color_u)
            ax.set_title("Kernel Density of Propensity Scores")

        elif to_plot == 'outcome':
            v = self.QAW - self.df[self.outcome]
            ax = plot_kde_accuracy(values=v.dropna(), bw_method=bw_method, fill=fill, color=color)
            ax.set_title("Kernel Density of Accuracy")

        else:
            raise ValueError("Please use one of the following options for `to_plot`; 'treatment', 'outcome'")

        return ax

    def plot_love(self, color_unweighted='r', color_weighted='b', shape_unweighted='o', shape_weighted='o'):
        """Generates a Love-plot to detail covariate balance based on the IPTW weights. Further details on the usage of
        this plot are available in Austin PC & Stuart EA 2015 https://onlinelibrary.wiley.com/doi/full/10.1002/sim.6607

        The Love plot generates a dashed line at standardized mean difference of 0.10. Ideally, weighted SMD are below
        this level. Below 0.20 may also be sufficient. Variables above this level may be unbalanced despite the
        weighting procedure. Different functional forms (or approaches like machine learning) may be worth considering

        Parameters
        ----------
        color_unweighted : str, optional
            Color for the unweighted standardized mean differences. Default is red
        color_weighted : str, optional
            Color for the weighted standardized mean differences. Default is blue
        shape_unweighted : str, optional
            Shape of points for the unweighted standardized mean differences. Default is circles
        shape_weighted:
            Shape of points for the weighted standardized mean differences. Default is circles

        Returns
        -------
        axes
            Matplotlib axes of the Love plot
        """
        if not self._fit_outcome_model or not self._fit_exposure_model:
            raise ValueError("The exposure_model and outcome_model function must be ran before any diagnostics")

        if self._fit_missing_model:
            ps = self.g1W * np.where(self.df[self.exposure] == 1, self.m1W, self.m0W)
        else:
            ps = self.g1W
        df = self.df.copy()
        df['_g1_'] = ps
        df['_ipw_'] = np.where(df[self.exposure] == 1, 1 / df['_g1_'], 1 / (1 - df['_g1_']))
        ax = plot_love(df=df, treatment=self.exposure, weight='_ipw_', formula=self.__mweight,
                       color_unweighted=color_unweighted, color_weighted=color_weighted,
                       shape_unweighted=shape_unweighted, shape_weighted=shape_weighted)
        return ax


class StochasticTMLE:
    r"""Implementation of target maximum likelihood estimator for stochastic treatment plans. This implementation
    calculates TMLE for a time-fixed exposure and a single time-point outcome under a stochastic treatment plan of
    interest. By default, standard parametric regression models are used to calculate the estimate of interest. The
    StochasticTMLE estimator allows users to instead use machine learning algorithms from sklearn and PyGAM.

    Note
    ----
    Valid confidence intervals are only attainable with certain machine learning algorithms. These algorithms must be
    Donsker class for valid confidence intervals. GAM and LASSO are examples of alogorithms that are Donsker class

    Parameters
    ----------
    df : DataFrame
        Pandas dataframe containing the variables of interest
    exposure : str
        Column label for the exposure of interest
    outcome : str
        Column label for the outcome of interest
    alpha : float, optional
        Alpha for confidence interval level. Default is 0.05
    continuous_bound : float, optional
        Optional argument to control the bounding feature for continuous outcomes. The bounding process may result
        in values of 0,1 which are undefined for logit(x). This parameter adds or substracts from the scenarios of
        0,1 respectively. Default value is 0.0005
    verbose : bool, optional
        Optional argument for verbose estimation. With verbose estimation, the model fits for each result are printed
        to the console. It is highly recommended to turn this parameter to True when conducting model diagnostics

    Note
    ----
    TMLE is a doubly-robust substitution estimator. TMLE obtains the target estimate in a single step. The
    single-step TMLE is described further by van der Laan. For further details, see the listed references.

    Continuous outcomes must be bounded between 0 and 1. TMLE does this automatically for the user. Additionally,
    the average treatment effect is estimate is back converted to the original scale of Y. When scaling Y as Y*,
    some values may take the value of 0 or 1, which breaks a logit(Y*) transformation. To avoid this issue, Y* is
    bounded by the `continuous_bound` argument. The default is 0.0005, the same as R's tmle

    Following is a general narrative of the estimation procedure for TMLE with stochastic treatments

    1. Initial estimators for g-model (IPTW) and Q-model (g-formula) are fit. By default these estimators are based
    on parametric regression models. Additionally, machine learning algorithms can be used to estimate the g-model and
    Q-model.

    2. The auxiliary covariate is calculated (i.e. IPTW).

    .. math::

        H = \frac{p}{\widehat{\Pr}(A=a)}

    where `p` is the probability of treatment `a` under the stochastic intervention of interest.

    3. Targeting step occurs through estimation of `e` via a logistic regression model. Briefly a weighted logistic
    regression model (weighted by the auxiliary covariates) with the dependent variable as the observed outcome and
    an offset term of the Q-model predictions under the observed treatment (A).

    .. math::

        \text{logit}(Y) = \text{logit}(Q(A, W)) + \epsilon

    4. Stochastic interventions are evaluated through Monte Carlo integration for binary treatments. The different
    treatment plans are randomly applied and evaluated through the Q-model and then the targeting step via

    .. math::

        E[\text{logit}(Q(A=a, W)) + \hat{\epsilon}]

    This process is repeated a large number of times and the point estimate is the average of those individual treatment
    plans.

    Examples
    --------
    Setting up environment

    >>> from zepid import load_sample_data, spline
    >>> from zepid.causal.doublyrobust import StochasticTMLE
    >>> df = load_sample_data(False).dropna()
    >>> df[['cd4_rs1', 'cd4_rs2']] = spline(df, 'cd40', n_knots=3, term=2, restricted=True)

    Estimating TMLE for 0.2 being treated with ART

    >>> tmle = StochasticTMLE(df, exposure='art', outcome='dead')
    >>> tmle.exposure_model('male + age0 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
    >>> tmle.outcome_model('art + male + age0 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
    >>> tmle.fit(p=0.2)
    >>> tmle.summary()

    Estimating TMLE for conditional plan

    >>> tmle = StochasticTMLE(df, exposure='art', outcome='dead')
    >>> tmle.exposure_model('male + age0 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
    >>> tmle.outcome_model('art + male + age0 + cd40 + cd4_rs1 + cd4_rs2 + dvl0')
    >>> tmle.fit(p=[0.6, 0.4], conditional=["df['male']==1", "df['male']==0"])
    >>> tmle.summary()

    Estimating TMLE with machine learning algorithm from sklearn

    >>> from sklearn.linear_model import LogisticRegression
    >>> log1 = LogisticRegression(penalty='l1', random_state=201)
    >>> tmle = StochasticTMLE(df, 'art', 'dead')
    >>> tmle.exposure_model('male + age0 + cd40 + cd4_rs1 + cd4_rs2 + dvl0', custom_model=log1)
    >>> tmle.outcome_model('male + age0 + cd40 + cd4_rs1 + cd4_rs2 + dvl0', custom_model=log1)
    >>> tmle.fit(p=0.75)

    References
    ----------
    Muñoz ID, and Van Der Laan MJ. Population intervention causal effects based on stochastic interventions.
    Biometrics 68.2 (2012): 541-549.

    van der Laan MJ, and Sherri R. Targeted learning in data science: causal inference for complex longitudinal
    studies. Springer Science & Business Media, 2011.
    """
    def __init__(self, df, exposure, outcome, alpha=0.05, continuous_bound=0.0005, verbose=False):
        self.exposure = exposure
        self.outcome = outcome
        self._missing_indicator = '__missing_indicator__'
        self.df, self._miss_flag, self._continuous_outcome = check_input_data(data=df,
                                                                              exposure=exposure,
                                                                              outcome=outcome,
                                                                              estimator="StochasticTMLE",
                                                                              drop_censoring=True,
                                                                              drop_missing=True,
                                                                              binary_exposure_only=True)

        # Manage outcomes
        if self._continuous_outcome:
            self._continuous_min = np.min(self.df[outcome])
            self._continuous_max = np.max(self.df[outcome])
            self._cb = continuous_bound
            self.df[outcome] = tmle_unit_bounds(y=self.df[outcome], mini=self._continuous_min,
                                                maxi=self._continuous_max, bound=self._cb)
        else:
            self._cb = 0.0

        # Output attributes
        self.epsilon = None
        self.marginals_vector = None
        self.marginal_outcome = None
        self.alpha = alpha
        self.marginal_se = None
        self.marginal_ci = None
        self.conditional_se = None
        self.conditional_ci = None

        # Storage for items I need later
        self._outcome_model = None
        self._q_model = None
        self._Qinit_ = None
        self._treatment_model = None
        self._g_model = None
        self._resamples_ = None
        self._specified_bound_ = None
        self._denominator_ = None
        self._verbose_ = verbose
        self._out_model_custom = False
        self._exp_model_custom = False
        self._continuous_type = None

        # Custom model / machine learner storage
        self._g_custom_ = None
        self._q_custom_ = None

    def exposure_model(self, model, custom_model=None, bound=False):
        """Estimation of Pr(A=1|L), which is termed as g(A=1|L) in the literature. This value is used as the denominator
        for the inverse probability weights.

        Parameters
        ----------
        model : str
            Independent variables to predict the exposure. Example) 'var1 + var2 + var3'
        custom_model : optional
            Input for a custom model that is used in place of the logit model (default). The model must have the
            "fit()" and  "predict()" attributes. Both sklearn and supylearner are supported as custom models. In the
            background, TMLE will fit the custom model and generate the predicted probablities
        bound : float, list, optional
            Value between 0,1 to truncate predicted probabilities. Helps to avoid near positivity violations.
            Specifying this argument can improve finite sample performance for random positivity violations. However,
            truncating weights leads to additional confounding. Default is False, meaning no truncation of
            predicted probabilities occurs. Providing a single float assumes symmetric trunctation, where values below
            or above the threshold are set to the threshold value. Alternatively a list of floats can be provided for
            asymmetric trunctation, with the first value being the lower bound and the second being the upper bound
        """
        self._g_model = self.exposure + ' ~ ' + model

        if custom_model is None:  # Standard parametric regression model
            fitmodel = propensity_score(self.df, self._g_model, print_results=self._verbose_)
            pred = fitmodel.predict(self.df)
        else:  # User-specified prediction model
            # TODO need to create smart warning system  -- Issue #124
            self._exp_model_custom = True
            data = patsy.dmatrix(model + ' - 1', self.df)
            pred = exposure_machine_learner(xdata=np.asarray(data), ydata=np.asarray(self.df[self.exposure]),
                                            ml_model=custom_model, print_results=self._verbose_)

        if bound:  # Bounding predicted probabilities if requested
            pred2 = probability_bounds(pred, bounds=bound)
            self._specified_bound_ = np.sum(np.where(pred2 == pred, 0, 1))
            pred = pred2

        self._denominator_ = np.where(self.df[self.exposure] == 1, pred, 1 - pred)

    def outcome_model(self, model, custom_model=None, bound=False, continuous_distribution='gaussian'):
        """Estimation of E(Y|A,L), which is also written sometimes as Q(A,W) or Pr(Y=1|A,W).

        Parameters
        ----------
        model : str
            Independent variables to predict the exposure. Example) 'var1 + var2 + var3'
        custom_model : optional
            Input for a custom model that is used in place of the logit model (default). The model must have the
            "fit()" and  "predict()" attributes. Both sklearn and supylearner are supported as custom models. In the
            background, TMLE will fit the custom model and generate the predicted probablities
        bound : bool, optional
            This argument should ONLY be used if the outcome is continuous. Value between 0,1 to truncate the bounded
            predicted outcomes. Default is `False`, meaning no truncation of predicted outcomes occurs (unless a
            predicted outcome is outside the bounded continuous outcome). Providing a single float assumes symmetric
            trunctation. A list of floats can be provided for asymmetric trunctation.
        continuous_distribution : str, optional
            Distribution to use for continuous outcomes. Options are 'gaussian' for normal distributions and 'poisson'
            for Poisson distributions
        """
        self._q_model = self.outcome + ' ~ ' + model

        if custom_model is None:  # Standard parametric regression
            self._out_model_custom = False
            self._continuous_type = continuous_distribution
            if self._continuous_outcome:
                if (continuous_distribution.lower() == 'gaussian') or (continuous_distribution.lower() == 'normal'):
                    f = sm.families.family.Gaussian()
                elif continuous_distribution.lower() == 'poisson':
                    f = sm.families.family.Poisson()
                else:
                    raise ValueError("Only 'gaussian' and 'poisson' distributions are supported for continuous "
                                     "outcomes")
                self._outcome_model = smf.glm(self._q_model, self.df, family=f).fit()
            else:
                f = sm.families.family.Binomial()
                self._outcome_model = smf.glm(self._q_model, self.df, family=f).fit()
            if self._verbose_:
                print('==============================================================================')
                print('Outcome model')
                print(self._outcome_model.summary())
                print('==============================================================================')

            # Step 2) Estimation under the scenarios
            self._Qinit_ = self._outcome_model.predict(self.df)

        else:  # User-specified model
            # TODO need to create smart warning system -- Issue #124
            self._out_model_custom = True
            data = patsy.dmatrix(model + ' - 1', self.df)
            output = stochastic_outcome_machine_learner(xdata=np.asarray(data),
                                                        ydata=np.asarray(self.df[self.outcome]),
                                                        ml_model=custom_model,
                                                        continuous=self._continuous_outcome,
                                                        print_results=self._verbose_)
            self._Qinit_, self._outcome_model = output

        if not bound:  # Bounding predicted probabilities if requested
            bound = self._cb

        # This bounding step prevents continuous outcomes from being outside the range
        self._Qinit_ = probability_bounds(self._Qinit_, bounds=bound)

    def fit(self, p, conditional=None, samples=100, seed=None):
        """Calculate the effect from the predicted exposure probabilities and predicted outcome values using the TMLE
        procedure. Confidence intervals are calculated using influence curves.

        Parameters
        ----------
        p : float, list, tuple
            Proportion that correspond to the number of persons treated (all values must be between 0.0 and 1.0). If
            conditional is specified, p must be a list/tuple of floats of the same length
        conditional : None, list, tuple, optional
            A
        samples : int, optional
            Number of samples to use for the Monte Carlo integration procedure
        seed : None, int, optional
            Seed for the Monte Carlo integration procedure

        Note
        ----
        Exposure and outcome models must be specified prior to `fit()`

        Returns
        -------
        `StochasticTMLE` gains `marginal_vector` and `marginal_outcome` along with `marginal_ci`
        """
        if self._denominator_ is None:
            raise ValueError("The exposure_model() function must be specified before the fit() function")
        if self._Qinit_ is None:
            raise ValueError("The outcome_model() function must be specified before the fit() function")

        if seed is None:
            pass
        else:
            np.random.seed(seed)

        p = np.array(p)
        if np.any(p > 1) or np.any(p < 0):
            raise ValueError("All specified treatment probabilities must be between 0 and 1")
        if conditional is not None:
            if len(p) != len(conditional):
                raise ValueError("'p' and 'conditional' must be the same length")

        # Step 1) Calculating clever covariate (HAW)
        if conditional is None:
            numerator = np.where(self.df[self.exposure] == 1, p, 1 - p)
        else:
            df = self.df.copy()
            stochastic_check_conditional(df=self.df, conditional=conditional)
            numerator = np.array([np.nan] for i in range(self.df.shape[0]))
            for c, prop in zip(conditional, p):
                numerator = np.where(eval(c), np.where(df[self.exposure] == 1, prop, 1 - prop), numerator)

        haw = np.array(numerator / self._denominator_).astype(float)

        # Step 2) Estimate from Q-model
        # process completed in outcome_model() function and stored in self._Qinit_

        # Step 3) Target parameter TMLE
        self.epsilon = self.targeting_step(y=self.df[self.outcome], q_init=self._Qinit_, iptw=haw,
                                           verbose=self._verbose_)

        # Step 4) Monte-Carlo Integration procedure
        q_star_list = []
        q_i_star_list = []
        self._resamples_ = samples
        for i in range(samples):
            # Applying treatment plan
            df = self.df.copy()
            if conditional is None:
                df[self.exposure] = np.random.binomial(n=1, p=p, size=df.shape[0])
            else:
                df[self.exposure] = np.nan
                for c, prop in zip(conditional, p):
                    df[self.exposure] = np.random.binomial(n=1, p=prop, size=df.shape[0])

            # Outcome model under treatment plan
            if self._out_model_custom:
                _, data_star = patsy.dmatrices(self._q_model + ' - 1', self.df)
                y_star = stochastic_outcome_predict(xdata=data_star,
                                                    fit_ml_model=self._outcome_model,
                                                    continuous=self._continuous_outcome)
            else:
                y_star = self._outcome_model.predict(df)

            # Targeted Estimate
            logit_qstar = np.log(probability_to_odds(y_star)) + self.epsilon  # logit(Y^*) + e
            q_star = odds_to_probability(np.exp(logit_qstar))  # Y^*
            q_i_star_list.append(q_star)  # Saving Y_i^* for marginal variance
            q_star_list.append(np.mean(q_star))  # Saving E[Y^*]

        if self._continuous_outcome:
            self.marginals_vector = tmle_unit_unbound(np.array(q_star_list),
                                                      mini=self._continuous_min,
                                                      maxi=self._continuous_max)
            y_ = np.array(tmle_unit_unbound(self.df[self.outcome],
                                            mini=self._continuous_min,
                                            maxi=self._continuous_max))
            yq0_ = tmle_unit_unbound(self._Qinit_,
                                     mini=self._continuous_min,
                                     maxi=self._continuous_max)
            yqstar_ = tmle_unit_unbound(np.array(q_i_star_list),
                                        mini=self._continuous_min,
                                        maxi=self._continuous_max)

        else:
            self.marginals_vector = q_star_list
            y_ = np.array(self.df[self.outcome])
            yq0_ = self._Qinit_
            yqstar_ = np.array(q_i_star_list)

        self.marginal_outcome = np.mean(self.marginals_vector)

        # Step 5) Estimating Var(psi)
        zalpha = norm.ppf(1 - self.alpha / 2, loc=0, scale=1)

        # Marginal variance estimator
        variance_marginal = self.est_marginal_variance(haw=haw, y_obs=y_, y_pred=yq0_,
                                                       y_pred_targeted=np.mean(yqstar_, axis=0),
                                                       psi=self.marginal_outcome)
        self.marginal_se = np.sqrt(variance_marginal) / np.sqrt(self.df.shape[0])
        self.marginal_ci = [self.marginal_outcome - zalpha * self.marginal_se,
                            self.marginal_outcome + zalpha * self.marginal_se]

        # Conditional on W variance estimator (not generally recommended but I need it for other work)
        variance_conditional = self.est_conditional_variance(haw=haw, y_obs=y_, y_pred=yq0_)
        self.conditional_se = np.sqrt(variance_conditional) / np.sqrt(self.df.shape[0])
        self.conditional_ci = [self.marginal_outcome - zalpha * self.conditional_se,
                               self.marginal_outcome + zalpha * self.conditional_se]

    def summary(self, decimal=3):
        """Prints summary of the estimated incidence under the specified treatment plan

        Parameters
        ----------
        decimal : int, optional
            Number of decimal places to display. Default is 3
        """
        if self.marginal_outcome is None:
            raise ValueError('The fit() statement must be ran before summary()')

        print('======================================================================')
        print('          Stochastic Targeted Maximum Likelihood Estimator            ')
        print('======================================================================')

        fmt = 'Treatment:        {:<15} No. Observations:     {:<20}'
        print(fmt.format(self.exposure, self.df.shape[0]))
        fmt = 'Outcome:          {:<15} No. Truncated:        {:<20}'
        if self._specified_bound_ is None:
            b = 0
        else:
            b = self._specified_bound_
        print(fmt.format(self.outcome, b))
        fmt = 'Q-Model:          {:<15} g-model:              {:<20}'
        print(fmt.format('Logistic', 'Logistic'))
        fmt = 'No. Resamples:    {:<15}'
        print(fmt.format(self._resamples_))

        print('======================================================================')
        print('Overall incidence:      ', np.round(self.marginal_outcome, decimals=decimal))
        print('======================================================================')
        print('Marginal 95% CL:        ', np.round(self.marginal_ci, decimals=decimal))
        print('Conditional 95% CL:     ', np.round(self.conditional_ci, decimals=decimal))
        print('======================================================================')

    def run_diagnostics(self, decimal=3):
        """Provides some summary diagnostics for `StochasticTMLE`. Diagnostics must be called are the fit() function is
        called since weights are dependent on the specific treatment plan of interest

        Parameters
        ----------
        decimal : int, optional
            Number of decimal places to display. Default is 3
        """
        if self.marginal_outcome is None:
            raise ValueError('The fit() statement must be ran before summary()')

        print('======================================================================')
        print('                         Diagnostics                                  ')
        print('======================================================================')
        print('                 Natural Course Prediction Accuracy                   ')
        value = self.df[self.outcome] - self._Qinit_
        fmt = 'Mean:                    {:<20}'
        print(fmt.format(np.round(np.mean(value), decimals=decimal)))
        fmt = 'Standard Deviation:      {:<20}'
        print(fmt.format(np.round(np.std(value, ddof=1), decimals=decimal)))
        fmt = 'Median:                  {:<20}'
        print(fmt.format(np.round(np.median(value), decimals=decimal)))
        fmt = 'Minimum value:           {:<20}'
        print(fmt.format(np.round(np.min(value), decimals=decimal)))
        fmt = 'Maximum value:           {:<20}'
        print(fmt.format(np.round(np.max(value), decimals=decimal)))

        print('----------------------------------------------------------------------')
        print('               Inverse Probability Weight Denominator                 ')
        w = 1 / self._denominator_
        fmt = 'Mean (expected=2):       {:<20}'
        print(fmt.format(np.round(np.mean(w), decimals=decimal)))
        fmt = 'Standard Deviation:      {:<20}'
        print(fmt.format(np.round(np.std(w, ddof=1), decimals=decimal)))
        fmt = 'Minimum value:           {:<20}'
        print(fmt.format(np.round(np.min(w), decimals=decimal)))
        fmt = 'Maximum value:           {:<20}'
        print(fmt.format(np.round(np.max(w), decimals=decimal)))

        print('----------------------------------------------------------------------')
        print('                        Targeting Step                                ')
        fmt = 'Epsilon:                 {:<20}'
        print(fmt.format(np.round(self.epsilon, decimals=decimal)))

        print('======================================================================')

        # Generating plots
        df = self.df.copy()
        df['_g1_'] = np.where(df[self.exposure] == 1, self._denominator_, 1 - self._denominator_)

        plt.figure(figsize=[8, 4])
        plt.subplot(121)
        plot_kde(df=df, treatment=self.exposure, probability='_g1_')
        plt.title("Kernel Density of Propensity Scores")

        plt.subplot(122)
        plot_kde_accuracy(values=value, color='green')
        plt.title("Kernel Density of Accuracy")
        plt.tight_layout()
        plt.show()

    @staticmethod
    def targeting_step(y, q_init, iptw, verbose):
        f = sm.families.family.Binomial()
        log = sm.GLM(y,  # Outcome / dependent variable
                     np.repeat(1, y.shape[0]),  # Generating intercept only model
                     offset=np.log(probability_to_odds(q_init)),  # Offset by g-formula predictions
                     freq_weights=iptw,  # Weighted by calculated IPW
                     family=f).fit()

        if verbose:  # Optional argument to print each intermediary result
            print('==============================================================================')
            print('Targeting Model')
            print(log.summary())

        return log.params[0]  # Returns single-step estimated Epsilon term

    @staticmethod
    def est_marginal_variance(haw, y_obs, y_pred, y_pred_targeted, psi):
        # https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4965321/
        doqg_psi_sq = (haw*(y_obs - y_pred) + y_pred_targeted - psi)**2
        var_est = np.mean(doqg_psi_sq)
        return var_est

    @staticmethod
    def est_conditional_variance(haw, y_obs, y_pred):
        # https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4965321/
        doqg_psi_sq = (haw*(y_obs - y_pred))**2
        var_est = np.mean(doqg_psi_sq)
        return var_est



import warnings
import numpy as np
import pandas as pd
import statsmodels.api as sm
import statsmodels.formula.api as smf
from statsmodels.genmod.families import links


class MonteCarloGFormula:
    r"""Time-varying implementation of the Monte Carlo g-formula. The Monte Carlo estimator is useful for survival data.
    For an extensive walkthrough of the Monte Carlo g-formula, see Keil et al. 2014 and other listed references. This
    implementation has four options for the treatment courses:

    Options for treatments
    * all : all individuals are given treatment
    * none : no individuals are given treatment
    * natural : individuals retain their observed treatment
    * custom : create a custom treatment. When specifying this, the dataframe must be  referred to as 'g' The
        following is an example that selects those whose age is 25 or older and are females
        Ex) treatment="((g['age0']>=25) & (g['male']==0))

    Note
    ----
    Custom treatments use a "magic-g" parameter. Internally, the g-formula implementation names the data set as `g`.
    Therefore, when using custom treatment specifications, the data set must be referred to as `g` when following
    the pandas selection syntax

    Currently, only binary exposures and a binary outcomes are supported. Logistic regression models are used to
    predict exposures and outcomes via statsmodels.
    See http://zepid.readthedocs.io/en/latest/ for an example (highly recommended)

    Parameters
    ----------
    df : DataFrame
        Pandas dataframe containing the variables of interest
    idvar : str
        ID column label
    exposure : str
        Treatment column label
    outcome : str
        Outcome column label
    time_out : str
        End of follow-up period time column label
    time_in : str
        Start of follow-up period time label
    weights : str, optional
        Column label for weights. Default is None, which assumes every observations has the same weight (i.e. 1)

    Notes
    -----
    1) Monte Carlo increases by time units of one. Input dataset should reflect this
    2) Only binary exposures and binary outcomes are supported
    3) Binary and continuous covariates are supported
    4) The labeling of the covariate models is important. They are fit in the order that they are labeled!
    5) Fit the natural course model first and compare to the observed data. They should be similar

    Process for the Monte Carlo g-formula

    1) run lines in "in_recode"
    2) time-varying covariates, order ascending in from "labels"

        a) predict time-varying covariate
        b) run lines in "recode" from "add_covariate_model()"

    3) predict exposure / apply exposure pattern
    4) predict outcome
    5) run lines in "out_recode"
    6) lag variables in "lags"
    7) append current-time rows to full dataframe
    8) Repeat till t_max is met

    Examples
    --------
    Setting up the environment

    >>> import numpy as np
    >>> from zepid import load_sample_data, spline
    >>> from zepid.causal.gformula import MonteCarloGFormula
    >>> df = load_sample_data(timevary=True)
    >>> df['lag_art'] = df['art'].shift(1)
    >>> df['lag_art'] = np.where(df.groupby('id').cumcount() == 0, 0, df['lag_art'])
    >>> df['lag_cd4'] = df['cd4'].shift(1)
    >>> df['lag_cd4'] = np.where(df.groupby('id').cumcount() == 0, df['cd40'], df['lag_cd4'])
    >>> df['lag_dvl'] = df['dvl'].shift(1)
    >>> df['lag_dvl'] = np.where(df.groupby('id').cumcount() == 0, df['dvl0'], df['lag_dvl'])
    >>> df[['age_rs0', 'age_rs1', 'age_rs2']] = spline(df, 'age0', n_knots=4, term=2, restricted=True)
    >>> df['cd40_sq'] = df['cd40'] ** 2
    >>> df['cd40_cu'] = df['cd40'] ** 3
    >>> df['cd4_sq'] = df['cd4'] ** 2
    >>> df['cd4_cu'] = df['cd4'] ** 3
    >>> df['enter_sq'] = df['enter'] ** 2
    >>> df['enter_cu'] = df['enter'] ** 3

    Estimating the g-formula with the Monte Carlo estimator

    >>> g = MonteCarloGFormula(df, idvar='id', exposure='art', outcome='dead', time_in='enter', time_out='out')

    >>> # Specifying the exposure/treatment model
    >>> exp_m = 'male + age0 + age_rs0 + age_rs1 + age_rs2 + cd40 + cd40_sq + cd40_cu + dvl0 + cd4 + cd4_sq +' +
    >>>         'cd4_cu + dvl + enter + enter_sq + enter_cu'
    >>> g.exposure_model(exp_m, restriction="g['lag_art']==0")  # restriction enforces intent-to-treat

    >>> # Specifying the outcome model
    >>> out_m = 'art + male + age0 + age_rs0 + age_rs1 + age_rs2 + cd40 + cd40_sq + cd40_cu + dvl0 + cd4 +' +
    >>>         'cd4_sq + cd4_cu + dvl + enter + enter_sq + enter_cu'
    >>> g.outcome_model(out_m, restriction="g['drop']==0")  # restriction enforces loss-to-follow-up

    >>> # Specifying the time-varying confounder models
    >>> dvl_m = 'male + age0 + age_rs0 + age_rs1 + age_rs2 + cd40 + cd40_sq + cd40_cu + dvl0 + lag_cd4 + ' +
    >>>         'lag_dvl + lag_art + enter + enter_sq + enter_cu'
    >>> g.add_covariate_model(label=1, covariate='dvl', model=dvl_m, var_type='binary')
    >>> cd4_m = 'male + age0 + age_rs0 + age_rs1 + age_rs2 +  cd40 + cd40_sq + cd40_cu + dvl0 + lag_cd4 + ' +
    >>>         'lag_dvl + lag_art + enter + enter_sq + enter_cu'
    >>> cd4_recode_scheme = ("g['cd4'] = np.maximum(g['cd4'],1);"  # Recode scheme makes sure variables are recoded
    >>>                      "g['cd4_sq'] = g['cd4']**2;"
    >>>                      "g['cd4_cu'] = g['cd4']**3")
    >>> g.add_covariate_model(label=2, covariate='cd4', model=cd4_m, recode=cd4_recode_scheme, var_type='continuous')

    >>> # Specifying a model for informative censoring
    >>> cens_m = "male + age0 + age_rs0 + age_rs1 + age_rs2 +  cd40 + cd40_sq + cd40_cu + dvl0 + lag_cd4 +" +
    >>>          "lag_dvl + lag_art + enter + enter_sq + enter_cu"
    >>> g.censoring_model(cens_m)

    >>> # Estimating outcomes under a simulated Markov Chain Monte Carlo for natural course
    >>> g.fit(treatment="((g['art']==1) | (g['lag_art']==1))",  # Treatment plan (natural course in this case)
    >>>       lags={'art': 'lag_art',  # Creating variables to lag in the process
    >>>             'cd4': 'lag_cd4',
    >>>             'dvl': 'lag_dvl'},
    >>>       sample=50000,  # Number of resamples to use (should be large number to reduce simulation error)
    >>>       t_max=None,  # Maximum time to simulate to (None uses data set maximum time)
    >>>       in_recode=("g['enter_sq'] = g['enter']**2;"
    >>>                  "g['enter_cu'] = g['enter']**3"))  # How to recode time in each time-step
    >>> # See website documentation for further instructions
    >>> # (https://zepid.readthedocs.io/en/latest/Causal.html#g-computation-algorithm-monte-carlo)

    References
    ----------
    Keil, AP, Edwards, JK, Richardson, DB, Naimi, AI, Cole, SR (2014). The Parametric g-Formula for Time-
    to-Event Data: Intuition and a Worked Example. Epidemiology 25(6), 889-897
    """

    def __init__(self, df, idvar, exposure, outcome, time_in, time_out, weights=None):
        self.gf = df.sort_values(by=[idvar, time_out]).copy()
        self.idvar = idvar

        # Checking that treatment is binary
        if df[exposure].dropna().value_counts().index.isin([0, 1]).all():
            self.exposure = exposure
        else:
            raise ValueError('Only binary treatments/exposures are currently implemented')

        # Checking that outcome is binary
        self.outcome = outcome
        if df[outcome].dropna().value_counts().index.isin([0, 1]).all():
            self._competing_event = False
        else:
            raise ValueError("Only binary outcomes are currently supported")
            # self._competing_event = True

        # Generating an indicator for censoring
        self.gf['__uncensored__'] = np.where((self.gf[idvar] != self.gf[idvar].shift(-1)) & (self.gf[outcome] == 0),
                                             0, 1)
        self.gf['__uncensored__'] = np.where(self.gf[time_out] == np.max(df[time_out]),
                                             1, self.gf['__uncensored__'])

        self.exp_model = None
        self.out_model = None
        self.cens_model = None
        self.time_in = time_in
        self.time_out = time_out
        self._exposure_model_fit = False
        self._outcome_model_fit = False
        self._censor_model_fit = False
        self._covariate_models = []
        self._covariate_model_index = []
        self._covariate = []
        self._covariate_type = []
        self._covariate_recode = []
        self._weights = weights
        self.predicted_outcomes = None

    def exposure_model(self, model, restriction=None, print_results=True):
        """Add a specified regression model for the exposure. This is used for natural course estimation of the Monte
        Carlo g-formula. This must be specified before calling the fit function.

        Parameters
        ----------
        model : str
            Variables to include in the model for predicting the exposure. Must be contained within the input
            pandas dataframe when initialized. Format follows patsy standards
            For example) 'var1 + var2 + var3 + var4'
        restriction : str, optional
            Used to restrict the population that the regression model is fit to. Useful for Intent-to-Treat model
            fitting. The pandas dataframe must be referred to as 'g'. For example) "g['art']==1"
        print_results : bool, optional
            Whether to print the logistic regression model results to the terminal. Default is True
        """
        g = self.gf.copy()
        if restriction is not None:
            g = g.loc[eval(restriction)].copy()
        linkdist = sm.families.family.Binomial()

        if self._weights is None:  # Unweighted g-formula
            self.exp_model = smf.glm(self.exposure + ' ~ ' + model, g, family=linkdist).fit()
        else:  # Weighted g-formula
            self.exp_model = smf.glm(self.exposure + ' ~ ' + model, g, freq_weights=g[self._weights],
                                     family=linkdist).fit()

        if print_results:
            print('==============================================================================')
            print('Predict-Exposure Model')
            print(self.exp_model.summary())
            print('==============================================================================')
        self._exposure_model_fit = True

    def outcome_model(self, model, restriction=None, print_results=True):
        """Add a specified regression model for the outcome. Must be specified before the fit function.

        Parameters
        ----------
        model:
            Variables to include in the model for predicting the outcome. Must be contained within the input
            pandas dataframe when initialized. Format follows patsy standards
            For example) 'var1 + var2 + var3 + var4'
        restriction : str, optional
            Used to restrict the population that the regression model is fit to. Useful for Intent-to-Treat model
            fitting. The pandas dataframe must be referred to as 'g'. For example) "g['art']==1"
        print_results : bool, optional
            Whether to print the logistic regression model results to the terminal. Default is True
        """
        g = self.gf.copy()
        if restriction is not None:
            g = g.loc[eval(restriction)].copy()
        linkdist = sm.families.family.Binomial()

        if self._weights is None:  # Unweighted g-formula
            if self._competing_event:
                self.out_model = sm.MNLogit.from_formula(self.outcome + ' ~ ' + model, g).fit()
            else:
                self.out_model = smf.glm(self.outcome + ' ~ ' + model, g, family=linkdist).fit()

        else:  # Weighted g-formula
            if self._competing_event:
                raise ValueError("The weighted MonteCarloGFormula is not supported for competing events")
            self.out_model = smf.glm(self.outcome + ' ~ ' + model, g, freq_weights=g[self._weights],
                                     family=linkdist).fit()
        if print_results:
            print('==============================================================================')
            print('Outcome Model')
            print(self.out_model.summary())
            print('==============================================================================')
        self._outcome_model_fit = True

    def censoring_model(self, model, restriction=None, print_results=True):
        """Add a specified regression model for censoring. Specifying this model is optional, but is recommended when
        censoring occurs in your data set. Otherwise, you will be assuming non-informative censoring

        Parameters
        ----------
        model:
            Variables to include in the model for predicting the outcome. Must be contained within the input
            pandas dataframe when initialized. Format follows patsy standards
            For example) 'var1 + var2 + var3 + var4'
        restriction : str, optional
            Used to restrict the population that the regression model is fit to. Useful for Intent-to-Treat model
            fitting. The pandas dataframe must be referred to as 'g'. For example) "g['art']==1"
        print_results : bool, optional
            Whether to print the logistic regression model results to the terminal. Default is True
        """
        g = self.gf.copy()
        if restriction is not None:
            g = g.loc[eval(restriction)].copy()
        linkdist = sm.families.family.Binomial()

        if self._weights is None:  # Unweighted g-formula
            self.cens_model = smf.glm('__uncensored__ ~ ' + model, g, family=linkdist).fit()
        else:  # Weighted g-formula
            self.cens_model = smf.glm('__uncensored__ ~ ' + model, g, freq_weights=g[self._weights],
                                      family=linkdist).fit()
        if print_results:
            print('==============================================================================')
            print('Censoring Model')
            print(self.cens_model.summary())
            print('==============================================================================')
        self._censor_model_fit = True

    def add_covariate_model(self, label, covariate, model, restriction=None, recode=None, var_type='binary',
                            print_results=True):
        """Add a specified regression model for time-varying confounders. Unlike the exposure and outcome models, a
        covariate model does NOT have to be specified. Additionally, *n* covariate models can be specified for *n*
        time-varying covariates. Additional models are added by repeated calls for this function with the corresponding
        covariates and predictive regression equations

        Parameters
        ----------
        label : int
            Integer label for the covariate model. Covariate models are fit in ascending order within
             TimeVaryGFormula
        covariate : str
            Column label for time-varying confounder to be predicted
        model : str
            Variables to include in the model for predicting the outcome. Must be contained within the input
            pandas dataframe when initialized. Format follows patsy
            For example) 'var1 + var2 + var3 + var4'
        restriction : str, optional
            Used to restrict the population to fit the logistic regression model to. Useful for Intent-to-Treat
            model fitting. The pandas dataframe must be referred to as 'g'. For example) "g['art']==1"
        recode : str, optional
            This variable is vitally important for various functional forms implemented later in models. This
            is used to run some background code to recreate functional forms as the g-formula is estimated via fit()
            For an example, let's say we have age but we want the functional form to be quadratic. For this, we
            would set the recode="g['age_sq'] = g['age']**2;" Similar to TimeFixedGFormula, 'g' must be specified as the
            DataFrame object with the corresponding indexes. Also lines of executable code should end with ';', so
            Python knows that the line ends there. My apologies for this poor solution... I am working on a better way.
            In the background, Python executes the code input into recode
        var_type : str, optional
            Type of variable that the covariate is. Current options include 'binary' or 'continuous'
        print_results : bool, optional
            Whether to print the logistic regression model results to the terminal. Default is True
        """
        if type(label) is not int:
            raise ValueError('Label must be an integer')

        # Building predictive model
        g = self.gf.copy()
        if restriction is not None:
            g = g.loc[eval(restriction)].copy()

        if self._weights is None:  # Unweighted g-formula
            if var_type == 'binary':
                linkdist = sm.families.family.Binomial()
                m = smf.glm(covariate + ' ~ ' + model, g, family=linkdist)
            elif var_type == 'continuous':
                m = smf.gls(covariate + ' ~ ' + model, g)
            else:
                raise ValueError('Only binary or continuous covariates are currently supported')
        else:  # Weighted g-formula
            if var_type == 'binary':
                linkdist = sm.families.family.Binomial()
                m = smf.glm(covariate + ' ~ ' + model, g, freq_weights=g[self._weights], family=linkdist)
            elif var_type == 'continuous':
                m = smf.wls(covariate + ' ~ ' + model, g, weights=g[self._weights])
            else:
                raise ValueError('Only binary or continuous covariates are currently supported')

        f = m.fit()
        if print_results:
            print('==============================================================================')
            print('Covariate (' + str(covariate) + ') Model')
            print(f.summary())
            print('==============================================================================')

        # Adding to lists, it is used to predict variables later on for the time-varying...
        self._covariate_models.append(f)
        self._covariate_model_index.append(label)
        self._covariate.append(covariate)
        self._covariate_type.append(var_type)
        if recode is None:
            self._covariate_recode.append('None')  # Must be string for exec() to use later
        else:
            self._covariate_recode.append(recode)

    def fit(self, treatment, lags=None, sample=10000, t_max=None, in_recode=None, out_recode=None, low_memory=True):
        """Estimate the counterfactual outcomes under the specified treatment plan using the previously specified
        regression models. Both the exposure and outcome models need to be specified before fit can be called

        Parameters
        ----------
        treatment : str
            Treatment strategy. Options include
            * all : all individuals are given treatment
            * none : no individuals are given treatment
            * natural : individuals retain their observed treatment. Only available for MonteCarlo
            * custom : create a custom treatment. When specifying this, the dataframe must be  referred to as 'g'
                The following is an example that selects those whose age is 25 or older and are females
                Ex) treatment="((g['age0']>=25) & (g['male']==0))"
        lags : dict, optional
            Dictionary of variable names and the corresponding lagged variable name. This should be specified for all
            variables that are lagged. This parameter is only used for the Monte Carlo g-formula.
             As an example, {'art':'lagged_art'} would correctly lag the variable 'art' to be 'lagged_art' for each time
             step of the Monte Carlo procedure
        sample : int, optional
            Number of individuals to sample from the original data with replacement. This argument is only used by the
            Monte Carlo g-formula. The number of samples to use should be a large number to reduce simulation error.
            The default is 10000
        t_max : int, optional
            Maximum time to run Monte Carlo g-formula until. Default is None, which uses the maximum time of the input
            dataframe.
        in_recode : str, optional
            On the fly recoding of variables done before the Monte Carlo loop starts. Needed to do any kind of
            functional forms for entry times. This is executed at each start of the Monte Carlo g-formula time steps
        out_recode : str, optional
            On the fly recoding of variables done at the end of the Monte Carlo loop. Needed for operations like
            counting the number of days with a treatment. This is executed at each end of the Monte Carlo g-formula
            time steps
        low_memory : bool, optional
            This optional parameter controls whether the g-formula outputs a condensed data set or a data set containing
            each step of the Monte Carlo procedure. When outputting each step of the Monte Carlo procedure, this can
            be intensive on memory requirements. `low_memory` is set to `True` (the default) only the last observation
            per ID is added to the full data set.
            From the condensed data (`low_memory`), you are able to directly fit a Kaplan-Meier and estimate the effect
            of treatment.
            The full Monte Carlo data set is useful during the model building phase and checking whether models are
            correctly specified. It is less useful during estimation and bootstrapping confidence intervals
        """
        if self._outcome_model_fit is False:
            raise ValueError('Before the g-formula can be calculated, the outcome model must be specified')
        if not isinstance(treatment, str) and not isinstance(treatment, list):
            raise ValueError('Specified treatment must be a string object')
        if self._exposure_model_fit is False:
            raise ValueError('Before the g-formula can be calculated, the exposure model must be specified for '
                             'Monte Carlo estimation')

        # Monte Carlo Estimation
        # Re-sampling from data set for estimation procedure
        if self._weights is None:
            gs = self.gf.loc[(self.gf.groupby(self.idvar).cumcount() == 0) == True].sample(n=sample, replace=True)
        else:
            gs = self.gf.loc[(self.gf.groupby(self.idvar).cumcount() == 0) == True].sample(n=sample,
                                                                                           weights=self._weights,
                                                                                           replace=True)
        gs['uid_g_zepid'] = [v for v in range(sample)]

        # Background preparations
        gs[self.outcome] = 0
        mc_simulated_data = []
        g = gs.copy()
        g['uncensored'] = 1  # this tricks the Monte-Carlo procedure into not censoring anyone if cens_model None
        if t_max is None:  # getting maximum time steps to run g-formula
            t_max = np.max(self.gf[self.time_out])

        if len(self._covariate_models) != 0:
            cov_model_order = (sorted(range(len(self._covariate_model_index)),
                                      key=self._covariate_model_index.__getitem__))
            run_cov = True
        else:
            run_cov = False

        # Monte Carlo for loop
        for i in range(int(t_max)):
            g = g.loc[(g[self.outcome] == 0) & (g['uncensored'] == 1)].reset_index(drop=True).copy()
            g[self.time_in] = i
            if in_recode is not None:
                exec(in_recode)

            # predict time-varying covariates
            if run_cov:
                for j in cov_model_order:
                    g[self._covariate[j]] = self._predict(df=g, model=self._covariate_models[j],
                                                          variable=self._covariate_type[j])
                    exec(self._covariate_recode[j])

            # predict exposure when customized treatments
            if treatment == 'all':
                g[self.exposure] = 1
            elif treatment == 'none':
                g[self.exposure] = 0
            elif treatment == 'natural':
                g[self.exposure] = self._predict(df=g, model=self.exp_model, variable='binary')
            else:  # custom exposure pattern
                g[self.exposure] = self._predict(df=g, model=self.exp_model, variable='binary')
                g[self.exposure] = np.where(eval(treatment), 1, 0)

            # predict outcome
            if self._competing_event:
                # TODO placeholder for future addition of competing risks
                g[self.outcome] = self._predict(df=g, model=self.out_model, variable='multinomial')
            else:
                g[self.outcome] = self._predict(df=g, model=self.out_model, variable='binary')
            g[self.time_out] = i + 1

            # predict censoring
            if self._censor_model_fit:
                g['uncensored'] = self._predict(df=g, model=self.cens_model, variable='binary')
                g[self.outcome] = np.where(g['uncensored'] == 1, g[self.outcome], 0)

            # last iteration, marking everyone as censored
            if i == t_max - 1:
                g['uncensored'] = 0

            # executing any code before appending
            if out_recode is not None:
                exec(out_recode)

            # updating lagged variables
            if lags is not None:
                for k, v in lags.items():
                    g[v] = g[k]

            # stacking simulated data in a list
            if low_memory:  # Only stacking when censored or failed
                mc_simulated_data.append(g.loc[(g[self.outcome] > 0) | (g['uncensored'] == 0)])
            else:  # Stacks all simulated observations
                mc_simulated_data.append(g)

        try:
            gs = pd.concat(mc_simulated_data, ignore_index=True, sort=False)
        except TypeError:  # gets around pandas <0.22 error
            gs = pd.concat(mc_simulated_data, ignore_index=True)

        cols_to_keep = ['uid_g_zepid', self.idvar, self.exposure, self.outcome, self.time_in,
                        self.time_out] + self._covariate
        self.predicted_outcomes = gs[cols_to_keep].sort_values(by=['uid_g_zepid', self.time_in]).reset_index(drop=True)

    @staticmethod
    def _predict(df, model, variable):
        """Hidden predict method to shorten the Monte Carlo estimation code
        """
        pp = model.predict(df)
        if variable == 'binary':
            pred = np.random.binomial(1, pp, size=len(pp))
        elif variable == 'continuous':
            pred = np.random.normal(loc=pp, scale=np.std(model.resid), size=len(pp))
        elif variable == 'multinomial':
            # Approach: 1
            # pred_cols = np.random.multinomial(n=1, pvals=pp, size=pp.shape[0])
            # pred = np.zeros(pp.shape[0])
            # for i, j in zip(range(pred_cols.shape[1]), pp.columns):
            #     pred = np.where(pred_cols[:, i], j, pred)
            # Approach: 2
            # pred = np.random.choice(a=list(pp.columns), p=pp, size=pp.shape[0])
            raise ValueError('That option is not supported')
        else:
            raise ValueError('That option is not supported')
        return pred

    @staticmethod
    def __predict__(df, model, variable, formula):
        """UNUSED FUNCTION

        New hidden predict method to shorten Monte Carlo estimation code. This is slower than statsmodels predict,
        but likely can be optimized. I would need to find a faster alternative to np.dot()...

        For this to work, I need to do the following:
            -need to have each ..._model() store a self.formula
            -switch all functions to sm.GLM or sm.GEE
        """
        import patsy
        from zepid.calc import odds_to_probability

        xdata = patsy.dmatrix(formula, df)  # , return_type='dataframe')
        # pred = xdata.mul(np.array(model.params), axis='columns').sum(axis=1)
        pred = xdata.dot(model.params)  # TODO optimize this...

        if variable == 'binary':
            pred = np.random.binomial(1, odds_to_probability(np.exp(pred)), size=xdata.shape[0])
        elif variable == 'continuous':
            pred = np.random.normal(loc=pred, scale=np.std(model.resid), size=len(pp))
        # TODO add optimization for multinomial (if applicable)
        else:
            raise ValueError('That option is not supported')
        return pred


class IterativeCondGFormula:
    r"""Iterative conditional g-formula estimator. This time-varying parametric g-formula uses the iterative conditional
    approach (also referred to as the sequential regression). The iterative conditional estimator is useful for
    longitudinal data and requires less model specification than the Monte Carlo g-formula. The iterative conditional
    uses a mathematical trick to estimate the marginal outcome distribution at the end of follow-up

    Unlike other implementations of the g-formula, the `IterativeCondGFormula` takes input data in a wide format.
    Additionally, treatments are specified by explicitly specifying the treatment plan array. See the examples for
    details

    Currently, only binary exposures and a binary outcomes are supported. Logistic regression models are used to
    predict exposures and outcomes via statsmodels. See Kreif et al. 2017 for a good description of the
    iterative conditional g-formula. See http://zepid.readthedocs.io/en/latest/ for an example

    Parameters
    ----------
    df : DataFrame
        Pandas dataframe containing the variables of interest
    exposures : list, array
        Treatment column label
    outcomes : list, array
        Outcome column label

    Notes
    -----
    Process for the sequential regression g-formula
    1) Identify individuals who followed the counterfactual treatment plan and had the outcome
    2) Fit a regression model for the outcome at time t for Y
    3) Predict outcomes under the observed treatment and the counterfactual treatment
    4) Repeat regression model fitting for t-1 to min(t)
    5) Take the mean predicted Y at the end to obtain the cumulative probability

    Examples
    --------
    Setting up the environment

    >>> from zepid import load_longitudinal_data
    >>> from zepid.causal.gformula import IterativeCondGFormula
    >>> df = load_longitudinal_data()

    Estimating the g-formula with the Monte Carlo estimator

    >>> icgf = IterativeCondGFormula(df, exposures=['A1', 'A2', 'A3'], outcomes=['Y1', 'Y2', 'Y3'])

    >>> # Specifying regression models for each treatment-outcome pair
    >>> icgf.outcome_model(models=['A1 + L1', 'A2 + A1 + L2', 'A3 + A2 + L3'], print_results=False)

    >>> # Estimating marginal 'Y3' under treat-all at every time
    >>> icgf.fit(treatments=[1, 1, 1])
    >>> print(icgf.marginal_outcome)

    >>> # Estimating marginal 'Y3' under treat-none at every time
    >>> icgf.fit(treatments=[0, 0, 0])
    >>> print(icgf.marginal_outcome)

    Custom treatments can be specified. Below is an example of treating everyone at the first and last time points

    >>> # Estimating marginal 'Y3' under custom treatment plan
    >>> icgf.fit(treatments=[1, 0, 1])
    >>> print(icgf.marginal_outcome)

    To estimate 'Y2', we can use a similar procedure but restrict our list of exposures and outcomes

    >>> icgf = IterativeCondGFormula(df, exposures=['A1', 'A2'], outcomes=['Y1', 'Y2'])
    >>> icgf.outcome_model(models=['A1 + L1', 'A2 + A1 + L2'], print_results=False)
    >>> icgf.fit(treatments=[1, 1])
    >>> print(icgf.marginal_outcome)

    References
    ----------
    Kreif, N., Tran, L., Grieve, R., De Stavola, B., Tasker, R. C., & Petersen, M. (2017). Estimating the
    comparative effectiveness of feeding interventions in the pediatric intensive care unit: a demonstration of
    longitudinal targeted maximum likelihood estimation. American Journal of Epidemiology, 186(12), 1370-1379.
    """

    def __init__(self, df, exposures, outcomes):
        self.gf = df.copy()

        # Check same number of treatments and outcomes
        if len(exposures) != len(outcomes):
            raise ValueError("The number of exposures must equal the number of outcomes")
        self.exposure = exposures

        # Checking that outcome is binary
        for o in outcomes:
            if not df[o].dropna().value_counts().index.isin([0, 1]).all():
                raise ValueError('Only binary outcomes are currently implemented')
        self.outcome = outcomes

        # Checking for recurrent outcomes. Recurrent are not currently supported
        # TODO add version to make missing data after observed outcome? Currently ignores, like R's ltmle
        if pd.Series(df[self.outcome].sum(axis=1, skipna=True) > 1).any():
            raise ValueError('Looks like your data has multiple outcomes. Recurrent outcomes are not currently '
                             'supported')

        self.exp_model = None
        self.out_model = None
        self._outcome_model_fit = False
        self.marginal_outcome = None
        self._modelform = None
        self._printseqregresults = None

    def outcome_model(self, models, print_results=True):
        """Add a specified regression model for the outcome. The number of regression models should match the number
        of outcomes specified. Must be specified before the fit function.

        Parameters
        ----------
        models: list, array
            Variables to include in the model for predicting the outcome. Must be contained within the input
            pandas dataframe when initialized. Format follows patsy standards
            For example) 'var1 + var2 + var3 + var4'
        print_results : bool, optional
            Whether to print the logistic regression model results to the terminal. Default is True
        """
        self._modelform = models
        self._printseqregresults = print_results
        self._outcome_model_fit = True

    def fit(self, treatments):
        """Estimate the counterfactual outcomes under the specified treatment plan using the previously specified
        regression models

        Parameters
        ----------
        treatments : list, array
            Treatment strategy. Options include
            * all : all individuals are given treatment
            * none : no individuals are given treatment
            * custom : create a custom treatment. When specifying this, the dataframe must be  referred to as 'g'
                The following is an example that selects those whose age is 25 or older and are females
                Ex) treatment="((g['age0']>=25) & (g['male']==0))
        """
        if self._outcome_model_fit is False:
            raise ValueError('Before the g-formula can be calculated, the outcome model must be specified')

        # Check array of treatments is either 1 row or same number of rows as input data
        treatment = np.array(treatments)
        if treatment.ndim == 1:
            treatment = np.tile(treatment, (self.gf[self.exposure].shape[0], 1))
        elif treatment.shape[1] == self.gf[self.exposure].shape[0]:
            pass
        else:
            raise ValueError("Specified treatments must be either a single row or have the same number of rows as the "
                             "input DataFrame")

        if len(self.exposure) != treatment.shape[1]:
            raise ValueError("The number of exposure variables and the number of treatments must be equal")

        linkdist = sm.families.family.Binomial()

        # Step 1: Creating indicator for individuals who followed counterfactual outcome
        # TODO remove this. Only needed for TMLE's IPTW calculation
        # adhere = self._identify_adherence_(observations=np.matrix(self.gf[self.exposure]), plan=treatment)
        # y_adhere = pd.DataFrame(self.gf[self.outcome].values * adhere.values,
        #                        columns=[self.outcome], index=self.gf.index)
        # y_adhere = y_adhere.where(~y_adhere.any(axis=1), y_adhere.fillna(1), axis=1)

        # Step 2: Sequential Regression Estimation
        treat_plan = ['_tplan_' + str(p) for p in range(treatment.shape[1])]
        df = self.gf.copy()
        df[treat_plan] = pd.DataFrame(treatment)
        # adhere_cols = list(y_adhere.columns)
        # df[adhere_cols] = y_adhere

        items_to_loop = zip(self.exposure[::-1], self.outcome[::-1], self._modelform[::-1],
                            treat_plan[::-1])  # , adhere_cols[::-1])
        for e, d, m, t in items_to_loop:
            # 2.1) Fit the model to the observed data
            if self.outcome[::-1].index(d) == 0:
                fm = smf.glm(d + ' ~ ' + m, df, family=linkdist).fit()  # GLM
            else:
                df[d] = np.where(df[prior_predict].isna(), df[d], df[prior_predict])
                fm = smf.glm(d + ' ~ ' + m, df, family=linkdist).fit()  # GLM
            if self._printseqregresults:
                print('==============================================================================')
                print('Sequential Outcome Model')
                print(fm.summary())
                print('==============================================================================')

            # 2.2) Generating predictions
            tf = df.copy()
            tf[self.exposure] = tf[treat_plan]
            prior_predict = '__pred_' + d
            df[prior_predict] = np.where(df[d].isna(), np.nan, fm.predict(tf))

            # 2.3) Carrying forward outcomes if followed counterfactual treatment and had outcome
            # TODO do I need this to do anything???
            # df['__cf_' + d] = np.where(df[f] == 1, 1, df['__pred_' + d])

        # Step 3) Calculating marginal outcome for that time period
        self.marginal_outcome = np.mean(df[prior_predict])  # Pulls the last time period

    @staticmethod
    def _identify_adherence_(observations, plan):
        """Background function to identify the adherence to the treatment plan. These individuals are special cases in
        the iterative conditional g-formula. They have their outcomes carried forward (since their counterfactual
        outcome under that treatment plan is that outcome
        """
        df = pd.DataFrame()
        for i in range(observations.shape[1]):
            # This code looks terrible due to numpy's matrix formulation and I don't have a better solution...
            df['_check_' + str(i)] = np.asarray(np.all(observations[:, 0:i + 1] == plan[:, 0:i + 1],
                                                       axis=1)).reshape(-1)

        return df



from .AIPW import AIPTW
from .TMLE import TMLE, StochasticTMLE
from .crossfit import (SingleCrossfitAIPTW, DoubleCrossfitAIPTW,
                       SingleCrossfitTMLE, DoubleCrossfitTMLE,
                       calculate_joint_estimate)



import warnings
import numpy as np
import pandas as pd
import statsmodels.api as sm
import statsmodels.formula.api as smf

from zepid.calc.utils import probability_bounds
from zepid.causal.utils import propensity_score, iptw_calculator


class IPSW:
    r"""Calculate inverse probability of sampling weights through logistic regression. Inverse probability of
    sampling weights are an extension of inverse probability weights to allow for the generalizability or the
    transportability of results.

    For generalizability, inverse probability of sampling weights take the following form

    .. math::

        IPSW = \frac{1}{\Pr(S=1|W)}

    where `W` is all the factors related to the sample selection process

    For transportability, the inverse probability of sampling weights are actually inverse odds of sampling weights.
    They take the following form

    .. math::

        IPSW = \frac{\Pr(S=0|W}{\Pr(S=1|W)}

    Confidence intervals should be obtained by using a non-parametric bootstrapping procedure

    Parameters
    ----------
    df : DataFrame
        Pandas dataframe containing all variables required for generalization/transportation. Should include
        all features related to sample selection, indicator for selection into the sample, and treatment/outcome
        information for the sample (selection == 1)
    exposure : str
        Column label for exposure/treatment of interest. Can be nan for all those not in sample
    outcome : str
        Column label for outcome of interest. Can be nan for all those not in sample
    selection : str
        Column label for indicator of selection into the sample. Should be 1 if individual comes from the study
        sample and 0 if individual is from random sample of source population
    generalize : bool, optional
        Whether the problem is a generalizability (True) problem or a transportability (False) problem. See notes
        for further details on the difference between the two estimation methods
    weights : None, str, optional
        For conditionally randomized trials, or observational research, inverse probability of treatment weights
        can be used to adjust for confounding. Before estimating the effect measures, this weight vector and the
        IPSW are multiplied to calculate new weights
        When weights is None, the data is assumed to come from a randomized trial, and does not need to be adjusted

    Note
    ----
    There are two related concepts; generalizability and transportability. Generalizability is when your study
    sample is part of your target population. For example, you want to generalize results from California to the
    entire United States. Transportability is when your study sample is not part of your target population. For
    example, we want to apply our results from California to Canada. Depending on the scenario, the IPSW have
    slightly different forms. `IPSW` allows for both of these problems

    Examples
    --------
    Setting up the environment

    >>> from zepid import load_generalize_data
    >>> from zepid.causal.generalize import IPSW
    >>> df = load_generalize_data(False)

    Generalizability for RCT results

    >>> ipsw = IPSW(df, exposure='A', outcome='Y', selection='S', generalize=True)
    >>> ipsw.sampling_model('L + W + L:W', print_results=False)
    >>> ipsw.fit()
    >>> ipsw.summary()

    Transportability for RCT results

    >>> ipsw = IPSW(df, exposure='A', outcome='Y', selection='S', generalize=False)
    >>> ipsw.sampling_model('L + W + L:W', print_results=False)
    >>> ipsw.fit()
    >>> ipsw.summary()

    For observational studies, IPTW can be used to account for confounders via the `treatment_model()` function

    >>> ipsw = IPSW(df, exposure='A', outcome='Y', selection='S')
    >>> ipsw.sampling_model('L + W + L:W', print_results=False)
    >>> ipsw.treatment_model('L', print_results=False)
    >>> ipsw.fit()
    >>> ipsw.summary()

    References
    ----------
    Lesko CR, Buchanan AL, Westreich D, Edwards JK, Hudgens MG, & Cole SR. (2017).
    Generalizing study results: a potential outcomes perspective. Epidemiology (Cambridge, Mass.), 28(4), 553.

    Westreich D, Edwards JK, Lesko CR, Stuart E, & Cole SR. (2017). Transportability of trial
    results using inverse odds of sampling weights. AJE, 186(8), 1010-1014.

    Dahabreh IJ, Robertson SE, Stuart EA, Hernan MA (2018). Transporting inferences from a
    randomized trial to a new target population. arXiv preprint arXiv:1805.00550.
    """

    def __init__(self, df, exposure, outcome, selection, generalize=True, weights=None):
        self.df = df.copy()
        self.sample = df.loc[df[selection] == 1].copy()
        self.target = df.loc[df[selection] == 0].copy()

        self.generalize = generalize  # determines whether IPSW or IOSW are calculated
        self.exposure = exposure
        self.outcome = outcome
        self.selection = selection
        self.weight = weights

        self.ipsw = None
        self.iptw = None
        self.risk_difference = None
        self.risk_ratio = None
        self._denominator_model = False

    def sampling_model(self, model_denominator, model_numerator='1', bound=None, stabilized=True, print_results=True):
        """Logistic regression model(s) for estimating sampling weights. The model denominator must be specified for
        both stabilized and unstabilized weights. The optional argument 'model_numerator' allows specification of the
        stabilization factor for the weight numerator. By default model results are returned

        Parameters
        ----------
        model_denominator : str
            String listing variables to predict the exposure, separated by +. For example, 'var1 + var2 + var3'. This
            is for the predicted probabilities of the denominator
        model_numerator : str, optional
            Optional string listing variables to predict the selection separated by +. Only used to calculate the
            numerator. Default ('1') calculates the overall probability of selection. In general, this is recommended.
            Adding in other variables means they are no longer accounted for in estimation of IPSW. Argument is also
            only used when calculating stabilized weights
        bound : float, list, optional
            Value between 0,1 to truncate predicted probabilities. Helps to avoid near positivity violations.
            Specifying this argument can improve finite sample performance for random positivity violations. However,
            inference becomes limited to the restricted population. Default is False, meaning no truncation of
            predicted probabilities occurs. Providing a single float assumes symmetric trunctation. A collection of
            floats can be provided for asymmetric trunctation
        stabilized : bool, optional
            Whether to generated stabilized IPSW. Default is True, which returns the stabilized IPSW
        print_results : bool, optional
            Whether to print the model results from the regression models. Default is True
        """
        if not stabilized:
            if model_numerator != '1':
                raise ValueError('Argument for model_numerator is only used for stabilized=True')

        dmodel = propensity_score(self.df, self.selection + ' ~ ' + model_denominator, print_results=print_results)

        self.sample['__denom__'] = dmodel.predict(self.sample)
        self._denominator_model = True

        # Stabilization factor if valid
        if stabilized:
            nmodel = propensity_score(self.df, self.selection + ' ~ ' + model_numerator, print_results=print_results)
            self.sample['__numer__'] = nmodel.predict(self.sample)
        else:
            self.sample['__numer__'] = 1

        if bound:
            self.sample['__denom__'] = probability_bounds(self.sample['__denom__'], bounds=bound)
            self.sample['__numer__'] = probability_bounds(self.sample['__numer__'], bounds=bound)

        # Calculate IPSW (generalizability)
        if self.generalize:
            self.sample['__ipsw__'] = self.sample['__numer__'] / self.sample['__denom__']

        # Calculate IOSW (transportability)
        else:
            if stabilized:
                self.sample['__ipsw__'] = (((1 - self.sample['__denom__']) / self.sample['__denom__']) *
                                           (self.sample['__numer__'] / (1 - self.sample['__numer__'])))
            else:
                self.sample['__ipsw__'] = (1 - self.sample['__denom__']) / self.sample['__denom__']

        self.ipsw = self.sample['__ipsw__']

    def treatment_model(self, model_denominator, model_numerator='1', bound=None, stabilized=True, print_results=True):
        """Logistic regression model(s) for estimating inverse probability of treatment weights (IPTW). The model
        denominator must be specified for both stabilized and unstabilized weights. The optional argument
        'model_numerator' allows specification of the stabilization factor for the weight numerator. By default model
        results are returned

        Parameters
        ----------
        model_denominator : str
            String listing variables to predict the exposure, separated by +. For example, 'var1 + var2 + var3'. This
            is for the predicted probabilities of the denominator
        model_numerator : str, optional
            Optional string listing variables to predict the selection separated by +. Only used to calculate the
            numerator. Default ('1') calculates the overall probability of selection. In general, this is recommended.
            Adding in other variables means they are no longer accounted for in estimation of IPSW. Argument is also
            only used when calculating stabilized weights
        bound : float, list, optional
            Value between 0,1 to truncate predicted probabilities. Helps to avoid near positivity violations.
            Specifying this argument can improve finite sample performance for random positivity violations. However,
            inference becomes limited to the restricted population. Default is False, meaning no truncation of
            predicted probabilities occurs. Providing a single float assumes symmetric trunctation. A collection of
            floats can be provided for asymmetric trunctation
        stabilized : bool, optional
            Whether to generated stabilized IPSW. Default is True, which returns the stabilized IPSW
        print_results : bool, optional
            Whether to print the model results from the regression models. Default is True
        """
        d, n, self.iptw = iptw_calculator(df=self.sample,
                                          treatment=self.exposure,
                                          model_denom=model_denominator, model_numer=model_numerator,
                                          weight=self.weight, stabilized=stabilized,
                                          standardize='population',
                                          bound=bound, print_results=print_results)

    # TODO add missing_model

    def fit(self):
        """Uses the calculated IPSW to obtain the risk difference and risk ratio from the sample. If weights are
        provided in the initial step, those weights are multiplied with IPSW to obtain the overall weights.

        The risk for the exposed and unexposed are calculated by taking the weighted averages.

        Returns
        -------
        `IPSW` gains `risk_difference` and `risk_ratio` which are the generalized risk difference and risk ratios for
        the exposure-outcome relationship based on the data and IPSW model
        """
        if not self._denominator_model:
            raise ValueError('The regression_models() function must be specified before effect measure estimation')

        if self.weight is not None:
            if self.iptw is None:
                self.sample['__ipw__'] = self.ipsw * self.sample[self.weight]
            else:
                self.sample['__ipw__'] = self.ipsw * self.iptw * self.sample[self.weight]
        else:
            if self.iptw is None:
                self.sample['__ipw__'] = self.ipsw
            else:
                self.sample['__ipw__'] = self.ipsw * self.iptw

        exp = self.sample[self.sample[self.exposure] == 1].copy()
        uxp = self.sample[self.sample[self.exposure] == 0].copy()

        r1 = np.average(exp[self.outcome], weights=exp['__ipw__'])
        r0 = np.average(uxp[self.outcome], weights=uxp['__ipw__'])

        self.risk_difference = r1 - r0
        self.risk_ratio = r1 / r0

    def summary(self, decimal=4):
        """Prints a summary of the results for the IPSW estimator

        Parameters
        ----------
        decimal : int, optional
            Number of decimal places to display in the result
        """
        print('======================================================================')
        if self.generalize:
            print('           Inverse Probability of Sampling Weights')
        else:
            print('               Inverse Odds of Sampling Weights')
        print('======================================================================')
        fmt = 'Treatment:        {:<15} Sample Observations:  {:<20}'
        print(fmt.format(self.exposure, self.sample.shape[0]))
        fmt = 'Outcome:          {:<15} Target Observations:  {:<20}'
        print(fmt.format(self.outcome, self.target.shape[0]))
        fmt = 'Target estimate:  {:<15} IP Treatment Weights: {:<20}'
        if self.generalize:
            g = 'Generalize'
        else:
            g = 'Transport'
        if self.iptw is None:
            w = 'No'
        else:
            w = 'Yes'
        print(fmt.format(g, w))

        print('----------------------------------------------------------------------')
        print('Risk Difference: ', round(float(self.risk_difference), decimal))
        print('Risk Ratio:      ', round(float(self.risk_ratio), decimal))
        print('======================================================================')


class GTransportFormula:
    r"""Calculate the g-transport-formula using a observed study sample and a sample from the target population.
    Broadly, the process for fitting the g-transport-formula is similar to the g-formula (as implemented in
    `TimeFixedGFormula`). Instead of predicting the potential outcomes of only the sample, the g-transport-formula
    predicts potential outcomes for the full target population

    For generalizability, we first fit a Q-model predicting the outcome as a function of the treatment and any
    modifiers (along with confounders if in observation data). Afterwards, we predict the potential outcomes for
    the entire population (S=1 and S=0). To obtain the marginal effect measure, we take the mean of the entire
    population (S=1 and S=0)

    For transportability, we similarly fit a Q-model in the observed sample and generate predictions for the entire
    sample. However, for transportability our sample is not part of the target population. Therefore, we only take
    the marginal of the S=0 group.

    Confidence intervals should be obtained by using a non-parametric bootstrapping procedure

    Parameters
    ----------
    df : DataFrame
        Pandas dataframe containing all variables required for generalization/transportation. Should include
        all features related to sample selection, indicator for selection into the sample, and treatment/outcome
        information for the sample (selection == 1)
    exposure : str
        Column label for exposure/treatment of interest. Can be nan for all those not in sample. Only binary
        exposures are currently supported
    outcome : str
        Column label for outcome of interest. Can be nan for all those not in sample
    selection : str
        Column label for indicator of selection into the sample. Should be 1 if individual comes from the study
        sample and 0 if individual is from random sample of source population
    outcome_type : str, optional
        Outcome variable type. Currently only 'binary', 'normal', and 'poisson variable types are supported
    generalize : bool, optional
        Whether the problem is a generalizability (True) problem or a transportability (False) problem. See notes
        for further details on the difference between the two estimation methods
    weights : None, str, optional
        Optional argument for weights. Can be used to input inverse probability of missing weights

    Note
    ----
    There are two related concepts; generalizability and transportability. Generalizability is when your study
    sample is part of your target population. For example, you want to generalize results from California to the
    entire United States. Transportability is when your study sample is not part of your target population. For
    example, we want to apply our results from California to Canada. Depending on the scenario, how the marginal
    risk difference is calculated is slightly different. `GTransportFormula` allows for both of these problems

    Examples
    --------
    Setting up the environment

    >>> from zepid import load_generalize_data
    >>> from zepid.causal.generalize import GTransportFormula
    >>> df = load_generalize_data(False)

    Generalizability

    >>> gtf = GTransportFormula(df, exposure='A', outcome='Y', selection='S', generalize=True)
    >>> gtf.outcome_model('A + L + L:A + W + W:A + W:A:L')
    >>> gtf.fit()
    >>> gtf.summary()

    Transportability

    >>> gtf = GTransportFormula(df, exposure='A', outcome='Y', selection='S', generalize=False)
    >>> gtf.outcome_model('A + L + L:A + W + W:A + W:A:L')
    >>> gtf.fit()
    >>> gtf.summary()

    For observational studies, confounders should be included in the Q-model

    References
    ----------
    Lesko CR, Buchanan AL, Westreich D, Edwards JK, Hudgens MG, & Cole SR. (2017).
    Generalizing study results: a potential outcomes perspective. Epidemiology (Cambridge, Mass.), 28(4), 553.

    Dahabreh IJ, Robertson SE, Stuart EA, Hernan MA (2018). Transporting inferences from a
    randomized trial to a new target population. arXiv preprint arXiv:1805.00550.
    """

    def __init__(self, df, exposure, outcome, selection, outcome_type='binary', generalize=True, weights=None):
        self.df = df.copy()
        self.sample = df.loc[df[selection] == 1].copy()
        self.target = df.loc[df[selection] == 0].copy()

        self.generalize = generalize  # determines the marginal calculation

        self.exposure = exposure
        self.outcome = outcome
        self.outcome_type = outcome_type
        self.selection = selection
        self.weight = weights

        self.risk_difference = None
        self.risk_ratio = None
        self.Weight = None
        self._outcome_model = False

    def outcome_model(self, model, print_results=True):
        """Build the model for the outcome. This is also referred to at the Q-model. This must be specified
        before the fit function. If it is not, an error will be raised.

        Parameters
        ----------
        model : str
            Variables to include in the model for predicting the outcome. Must be contained within the input
            pandas dataframe when initialized. Model form should contain the exposure, i.e. 'art + age + male'
        print_results : bool, optional
            Whether to print the logistic regression results to the terminal. Default is True
        """
        if self.exposure not in model:
            warnings.warn("It looks like '" + self.exposure + "' is not included in the outcome model.")

        if self.outcome_type == 'binary':
            linkdist = sm.families.family.Binomial()
        elif self.outcome_type == 'normal':
            linkdist = sm.families.family.Gaussian()
        elif self.outcome_type == 'poisson':
            linkdist = sm.families.family.Poisson()
        else:
            raise ValueError("Only 'binary', 'normal', and 'poisson' distributed outcomes are available")

        # Modeling the outcome
        if self.weight is None:
            m = smf.glm(self.outcome+' ~ '+model, self.sample, family=linkdist)
            self._outcome_model = m.fit()
        else:
            m = smf.glm(self.outcome+' ~ '+model, self.sample, family=linkdist,
                        freq_weights=self.sample[self.weight])
            self._outcome_model = m.fit()

        # Printing results of the model and if any observations were dropped
        if print_results:
            print('==============================================================================')
            print('Outcome Model')
            print(self._outcome_model.summary())
            print('==============================================================================')

    def fit(self):
        """Uses the g-transport formula to obtain the risk difference and risk ratio from the sample.

        Returns
        -------
        `GTransportFormula` gains `risk_difference` and `risk_ratio` which are the generalized risk difference and
        risk ratios for the exposure-outcome relationship
        """
        if not self._outcome_model:
            raise ValueError('The outcome_model() function must be specified before effect measure estimation')

        # Generalizability problem
        if self.generalize:
            dfa = self.df.copy()
            dfn = self.df.copy()
            dfa[self.exposure] = 1
            dfn[self.exposure] = 0

            ya = self._outcome_model.predict(dfa)
            yn = self._outcome_model.predict(dfn)

            if self.weight is not None:
                r1 = np.average(ya, weights=self.df[self.weight])
                r0 = np.average(yn, weights=self.df[self.weight])
            else:
                r1 = np.mean(ya)
                r0 = np.mean(yn)

        # Tranportability problem
        else:
            dfa = self.target.copy()
            dfn = self.target.copy()
            dfa[self.exposure] = 1
            dfn[self.exposure] = 0

            ya = self._outcome_model.predict(dfa)
            yn = self._outcome_model.predict(dfn)

            if self.weight is not None:
                r1 = np.average(ya, weights=self.target[self.weight])
                r0 = np.average(yn, weights=self.target[self.weight])
            else:
                r1 = np.mean(ya)
                r0 = np.mean(yn)

        self.risk_difference = r1 - r0
        self.risk_ratio = r1 / r0

    def summary(self, decimal=4):
        """Prints a summary of the results for the g-transport estimator

        Parameters
        ----------
        decimal : int, optional
            Number of decimal places to display in the result
        """
        print('======================================================================')
        print('                       g-Transport formula')
        print('======================================================================')
        fmt = 'Treatment:        {:<15} Sample Observations:  {:<20}'
        print(fmt.format(self.exposure, self.sample.shape[0]))
        fmt = 'Outcome:          {:<15} Target Observations:  {:<20}'
        print(fmt.format(self.outcome, self.target.shape[0]))
        fmt = 'Target estimate:  {:<15}'
        if self.generalize:
            print(fmt.format('Generalize'))
        else:
            print(fmt.format('Transport'))

        print('----------------------------------------------------------------------')
        print('Risk Difference: ', round(float(self.risk_difference), decimal))
        print('Risk Ratio:      ', round(float(self.risk_ratio), decimal))
        print('======================================================================')


class AIPSW:
    r"""Doubly robust estimator for generalizability. I haven't found a good name for it in the literature yet, so I
    am naming it augmented-IPSW (in honor of other doubly robust estimators like AIPTW and AIPMW).

    The process of estimating AIPSW follows other doubly robust estimators. We need to specify both the IPSW model
    and the g-transport model. From this information, the AIPSW is calculated via the following

    .. math::

        \psi = \frac{1}{n} \sum \left(E[Y|A=a,L,S=1] + \frac{I(S=1, A=a)}{\Pr(S=1|L)} (Y - E[Y|A=a,L,S=1])\right)

    For transportability problems, AIPSW takes the following form

    .. math::

        \psi = \frac{\sum IPSW\times I(S=1, A=a)(Y - E[Y|A=a,L,S=1]) + (1-S)E[Y|A=a,L,S=1]}{\Pr(S=0)}

    For generalizability, we first fit a Q-model predicting the outcome as a function of the treatment and any
    modifiers (along with confounders if in observation data). Next we calculate IPSW (with IPTW if there is any
    confounders). Afterwards, we predict the potential outcomes for the entire population (S=1 and S=0). We then
    use the above formula to calculate the marginal effect

    A similar process is done for transportability. Instead we merge g-transport and inverse odds of sampling weights

    Confidence intervals should be obtained by using a non-parametric bootstrapping procedure

    Parameters
    ----------
    df : DataFrame
        Pandas dataframe containing all variables required for generalization/transportation. Should include
        all features related to sample selection, indicator for selection into the sample, and treatment/outcome
        information for the sample (selection == 1)
    exposure : str
        Column label for exposure/treatment of interest. Can be nan for all those not in sample. Only binary
        exposures are currently supported
    outcome : str
        Column label for outcome of interest. Can be nan for all those not in sample
    selection : str
        Column label for indicator of selection into the sample. Should be 1 if individual comes from the study
        sample and 0 if individual is from random sample of source population
    generalize : bool, optional
        Whether the problem is a generalizability (True) problem or a transportability (False) problem. See notes
        for further details on the difference between the two estimation methods
    weights : None, str, optional
        Optional argument for weights. Can be used to input inverse probability of missing weights

    Note
    ----
    There are two related concepts; generalizability and transportability. Generalizability is when your study
    sample is part of your target population. For example, you want to generalize results from California to the
    entire United States. Transportability is when your study sample is not part of your target population. For
    example, we want to apply our results from California to Canada. Depending on the scenario, how the marginal
    risk difference is calculated is slightly different. `AIPSW` allows for both of these problems

    Examples
    --------
    Setting up the environment

    >>> from zepid import load_generalize_data
    >>> from zepid.causal.generalize import AIPSW
    >>> df = load_generalize_data(False)

    Generalizability for RCT

    >>> aipw = AIPSW(df, exposure='A', outcome='Y', selection='S', generalize=True)
    >>> aipw.sampling_model('L + W_sq', print_results=False)
    >>> aipw.outcome_model('A + L + L:A + W + W:A + W:A:L', print_results=False)
    >>> aipw.fit()
    >>> aipw.summary()

    Transportability for RCT

    >>> aipw = AIPSW(df, exposure='A', outcome='Y', selection='S', generalize=False)
    >>> aipw.sampling_model('L + W_sq', print_results=False)
    >>> aipw.outcome_model('A + L + L:A + W + W:A + W:A:L', print_results=False)
    >>> aipw.fit()
    >>> aipw.summary()

    Transportability for Observational study

    >>> df = load_generalize_data(True)
    >>> aipw = AIPSW(df, exposure='A', outcome='Y', selection='S', generalize=False)
    >>> aipw.sampling_model('L + W_sq', print_results=False)
    >>> aipw.treatment_model('L', print_results=False)
    >>> aipw.outcome_model('A + L + L:A + W + W:A + W:A:L', print_results=False)
    >>> aipw.fit()
    >>> aipw.summary()

    References
    ----------
    Dahabreh IJ, Robertson SE, Stuart EA, Hernan MA (2018). Transporting inferences from a
    randomized trial to a new target population. arXiv preprint arXiv:1805.00550.

    Dahabreh IJ, Hernan MA, Robertson SE, Buchanan A, Steingrimsson JA. (2019). Generalizing
    trial findings in nested trial designs with sub-sampling of non-randomized individuals. arXiv preprint
    arXiv:1902.06080.
    """

    def __init__(self, df, exposure, outcome, selection, generalize=True, weights=None):
        self.df = df.copy()
        self.sample = df[selection] == 1
        self.target = df[selection] == 0

        self.generalize = generalize  # determines whether IPSW or IOSW are calculated

        self.exposure = exposure
        self.outcome = outcome
        self.selection = selection
        if weights is not None:
            raise ValueError("AIPSW is not stable with `weights`. This functionality is currently unavailable")
        self.weight = weights

        self.ipsw = None
        self.iptw = None
        self._denominator_model = False
        self._outcome_model = False
        self._YA1 = None
        self._YA0 = None

        self.risk_difference = None
        self.risk_ratio = None

    def sampling_model(self, model_denominator, model_numerator='1', stabilized=True, print_results=True):
        """Logistic regression model(s) for estimating IPSW. The model denominator must be specified for both
        stabilized and unstabilized weights. The optional argument 'model_numerator' allows specification of the
        stabilization factor for the weight numerator. By default model results are returned

        Parameters
        ----------
        model_denominator : str
            String listing variables to predict the exposure, separated by +. For example, 'var1 + var2 + var3'. This
            is for the predicted probabilities of the denominator
        model_numerator : str, optional
            Optional string listing variables to predict the selection separated by +. Only used to calculate the
            numerator. Default ('1') calculates the overall probability of selection. In general, this is recommended.
            Adding in other variables means they are no longer accounted for in estimation of IPSW. Argument is also
            only used when calculating stabilized weights
        stabilized : bool, optional
            Whether to generated stabilized IPSW. Default is True, which returns the stabilized IPSW
        print_results : bool, optional
            Whether to print the model results from the regression models. Default is True
        """
        dmodel = propensity_score(self.df, self.selection + ' ~ ' + model_denominator, print_results=print_results)
        self.df['__denom__'] = dmodel.predict(self.df)
        self._denominator_model = True

        # Stabilization factor if valid
        if stabilized:
            nmodel = propensity_score(self.df, self.selection + ' ~ ' + model_numerator, print_results=print_results)
            self.df['__numer__'] = np.where(self.sample, nmodel.predict(self.df), 0)
        else:
            self.df['__numer__'] = np.where(self.sample, 1, 0)

        # Calculate IPSW (generalizability)
        if self.generalize:
            self.df['__ipsw__'] = self.df['__numer__'] / self.df['__denom__']

        # Calculate IOSW (transportability)
        else:
            if stabilized:
                self.df['__ipsw__'] = (((1 - self.df['__denom__']) / self.df['__denom__']) *
                                           (self.df['__numer__'] / (1 - self.df['__numer__'])))
            else:
                self.df['__ipsw__'] = (1 - self.df['__denom__']) / self.df['__denom__']

        self.ipsw = self.df['__ipsw__']

    def treatment_model(self, model_denominator, model_numerator='1', bound=None, stabilized=True, print_results=False):
        """Logistic regression model(s) for estimating inverse probability of treatment weights (IPTW). The model
        denominator must be specified for both stabilized and unstabilized weights. The optional argument
        'model_numerator' allows specification of the stabilization factor for the weight numerator. By default model
        results are returned

        Parameters
        ----------
        model_denominator : str
            String listing variables to predict the exposure, separated by +. For example, 'var1 + var2 + var3'. This
            is for the predicted probabilities of the denominator
        model_numerator : str, optional
            Optional string listing variables to predict the selection separated by +. Only used to calculate the
            numerator. Default ('1') calculates the overall probability of selection. In general, this is recommended.
            Adding in other variables means they are no longer accounted for in estimation of IPSW. Argument is also
            only used when calculating stabilized weights
        bound : float, list, optional
            Value between 0,1 to truncate predicted probabilities. Helps to avoid near positivity violations.
            Specifying this argument can improve finite sample performance for random positivity violations. However,
            inference becomes limited to the restricted population. Default is False, meaning no truncation of
            predicted probabilities occurs. Providing a single float assumes symmetric trunctation. A collection of
            floats can be provided for asymmetric trunctation
        stabilized : bool, optional
            Whether to generated stabilized IPTW. Default is True, which returns the stabilized IPTW
        print_results : bool, optional
            Whether to print the model results from the regression models. Default is True
        """
        d, n, self.iptw = iptw_calculator(df=self.df,
                                          treatment=self.exposure,
                                          model_denom=model_denominator, model_numer=model_numerator,
                                          weight=self.weight, stabilized=stabilized,
                                          standardize='population',
                                          bound=bound, print_results=print_results)

    def outcome_model(self, model, outcome_type='binary', print_results=True):
        """Build the g-transport model for the outcome. This is also referred to at the Q-model.

        Parameters
        ----------
        model : str
            Variables to include in the model for predicting the outcome. Must be contained within the input
            pandas dataframe when initialized. Model form should contain the exposure, i.e. 'art + age + male'
        outcome_type : str, optional
            Variable type for the outcome. Default is binary. Also supports 'normal' and 'poisson' distributed outcomes
        print_results : bool, optional
            Whether to print the logistic regression results to the terminal. Default is True
        """
        if self.exposure not in model:
            warnings.warn("It looks like '" + self.exposure + "' is not included in the outcome model.")

        if outcome_type == 'binary':
            linkdist = sm.families.family.Binomial()
        elif outcome_type == 'normal':
            linkdist = sm.families.family.Gaussian()
        elif outcome_type == 'poisson':
            linkdist = sm.families.family.Poisson()
        else:
            raise ValueError("Only 'binary', 'normal', and 'poisson' distributed outcomes are available")

        # Modeling the outcome
        df = self.df[self.sample].copy()

        if self.weight is None:
            m = smf.glm(self.outcome+' ~ '+model, df, family=linkdist)
            self._outcome_model = m.fit()
        else:
            m = smf.glm(self.outcome+' ~ '+model, df, family=linkdist, freq_weights=df[self.weight])
            self._outcome_model = m.fit()

        # Printing results of the model and if any observations were dropped
        if print_results:
            print('==============================================================================')
            print('Outcome Model')
            print(self._outcome_model.summary())
            print('==============================================================================')

        dfa = self.df.copy()
        dfa[self.exposure] = 1
        self._YA1 = self._outcome_model.predict(dfa)

        dfn = self.df.copy()
        dfn[self.exposure] = 0
        self._YA0 = self._outcome_model.predict(dfn)

    def fit(self):
        """Uses AIPSW formula to obtain the risk difference and risk ratio from the sample.

        Returns
        -------
        `AIPSW` gains `risk_difference` and `risk_ratio` which are the generalized risk difference and risk ratios for
        the exposure-outcome relationship based on the data and IPSW model
        """
        if not self._denominator_model:
            raise ValueError('sampling_model() function must be specified before effect measure estimation')
        if not self._outcome_model:
            raise ValueError('outcome_model() function must be specified before effect measure estimation')

        if self.weight is not None:
            if self.iptw is None:
                self.df['__ipw__'] = self.ipsw * self.df[self.weight]
            else:
                self.df['__ipw__'] = self.ipsw * self.iptw * self.df[self.weight]
        else:
            if self.iptw is None:
                self.df['__ipw__'] = self.ipsw
            else:
                self.df['__ipw__'] = self.ipsw * self.iptw

        # Generalizability problem
        if self.generalize:
            part1 = self._YA1
            part2 = np.where(self.sample & (self.df[self.exposure] == 1),
                             self.df['__ipw__'] * (self.df[self.outcome] - self._YA1), 0)
            r1 = np.mean(part1 + part2)

            part1 = self._YA0
            part2 = np.where(self.sample & (self.df[self.exposure] == 0),
                             self.df['__ipw__'] * (self.df[self.outcome] - self._YA0), 0)
            r0 = np.mean(part1 + part2)

        # Transportability problem
        else:
            part1 = np.where(self.sample & (self.df[self.exposure] == 1),
                             self.df['__ipw__'] * (self.df[self.outcome] - self._YA1), 0)
            part2 = (1 - self.df[self.selection]) * self._YA1
            r1 = np.sum(part1 + part2) / np.sum(1 - self.df[self.selection])

            part1 = np.where(self.sample & (self.df[self.exposure] == 0),
                             self.df['__ipw__'] * (self.df[self.outcome] - self._YA0), 0)
            part2 = (1 - self.df[self.selection]) * self._YA0
            r0 = np.sum(part1 + part2) / np.sum(1 - self.df[self.selection])

        self.risk_difference = r1 - r0
        self.risk_ratio = r1 / r0

    def summary(self, decimal=4):
        """Prints a summary of the results for the AIPSW estimator

        Parameters
        ----------
        decimal : int, optional
            Number of decimal places to display in the result
        """
        print('======================================================================')
        print('           Augmented Inverse Probability of Sampling Weights          ')
        print('======================================================================')
        fmt = 'Treatment:        {:<15} Sample Observations:  {:<20}'
        print(fmt.format(self.exposure, self.df[self.sample].shape[0]))
        fmt = 'Outcome:          {:<15} Target Observations:  {:<20}'
        print(fmt.format(self.outcome, self.df[self.target].shape[0]))
        fmt = 'Target estimate:  {:<15}'
        if self.generalize:
            print(fmt.format('Generalize'))
        else:
            print(fmt.format('Transport'))

        print('----------------------------------------------------------------------')
        print('Risk Difference: ', round(float(self.risk_difference), decimal))
        print('Risk Ratio:      ', round(float(self.risk_ratio), decimal))
        print('======================================================================')
